#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

from datetime import datetime, timezone

import pytest
from airbyte_cdk.sources.streams.concurrent.cursor import CursorField
from airbyte_cdk.sources.streams.concurrent.state_converters.abstract_stream_state_converter import ConcurrencyCompatibleStateType
from airbyte_cdk.sources.streams.concurrent.state_converters.datetime_stream_state_converter import (
    EpochValueConcurrentStreamStateConverter,
    IsoMillisConcurrentStreamStateConverter,
)


@pytest.mark.parametrize(
    "converter, input_state, is_compatible",
    [
        pytest.param(
            EpochValueConcurrentStreamStateConverter(),
            {"state_type": "date-range"},
            True,
            id="no-input-state-is-compatible-epoch",
        ),
        pytest.param(
            EpochValueConcurrentStreamStateConverter(),
            {
                "created_at": "2022_05_22",
                "state_type": ConcurrencyCompatibleStateType.date_range.value,
            },
            True,
            id="input-state-with-date_range-is-compatible-epoch",
        ),
        pytest.param(
            EpochValueConcurrentStreamStateConverter(),
            {
                "created_at": "2022_05_22",
                "state_type": "fake",
            },
            False,
            id="input-state-with-fake-state-type-is-not-compatible-epoch",
        ),
        pytest.param(
            EpochValueConcurrentStreamStateConverter(),
            {
                "created_at": "2022_05_22",
            },
            False,
            id="input-state-without-state_type-is-not-compatible-epoch",
        ),
        pytest.param(
            IsoMillisConcurrentStreamStateConverter(),
            {"state_type": "date-range"},
            True,
            id="no-input-state-is-compatible-isomillis",
        ),
        pytest.param(
            IsoMillisConcurrentStreamStateConverter(),
            {
                "created_at": "2022_05_22",
                "state_type": ConcurrencyCompatibleStateType.date_range.value,
            },
            True,
            id="input-state-with-date_range-is-compatible-isomillis",
        ),
        pytest.param(
            IsoMillisConcurrentStreamStateConverter(),
            {
                "created_at": "2022_05_22",
                "state_type": "fake",
            },
            False,
            id="input-state-with-fake-state-type-is-not-compatible-isomillis",
        ),
        pytest.param(
            IsoMillisConcurrentStreamStateConverter(),
            {
                "created_at": "2022_05_22",
            },
            False,
            id="input-state-without-state_type-is-not-compatible-isomillis",
        ),
    ],
)
def test_concurrent_stream_state_converter_is_state_message_compatible(converter, input_state, is_compatible):
    assert converter.is_state_message_compatible(input_state) == is_compatible


@pytest.mark.parametrize(
    "converter,start,state,expected_start",
    [
        pytest.param(
            EpochValueConcurrentStreamStateConverter(),
            None,
            {},
            EpochValueConcurrentStreamStateConverter().zero_value,
            id="epoch-converter-no-state-no-start-start-is-zero-value",
        ),
        pytest.param(
            EpochValueConcurrentStreamStateConverter(),
            1617030403,
            {},
            datetime(2021, 3, 29, 15, 6, 43, tzinfo=timezone.utc),
            id="epoch-converter-no-state-with-start-start-is-start",
        ),
        pytest.param(
            EpochValueConcurrentStreamStateConverter(),
            None,
            {"created_at": 1617030404},
            datetime(2021, 3, 29, 15, 6, 44, tzinfo=timezone.utc),
            id="epoch-converter-state-without-start-start-is-from-state",
        ),
        pytest.param(
            EpochValueConcurrentStreamStateConverter(),
            1617030404,
            {"created_at": 1617030403},
            datetime(2021, 3, 29, 15, 6, 44, tzinfo=timezone.utc),
            id="epoch-converter-state-before-start-start-is-start",
        ),
        pytest.param(
            EpochValueConcurrentStreamStateConverter(),
            1617030403,
            {"created_at": 1617030404},
            datetime(2021, 3, 29, 15, 6, 44, tzinfo=timezone.utc),
            id="epoch-converter-state-after-start-start-is-from-state",
        ),
        pytest.param(
            IsoMillisConcurrentStreamStateConverter(),
            None,
            {},
            IsoMillisConcurrentStreamStateConverter().zero_value,
            id="isomillis-converter-no-state-no-start-start-is-zero-value",
        ),
        pytest.param(
            IsoMillisConcurrentStreamStateConverter(),
            "2021-08-22T05:03:27.000Z",
            {},
            datetime(2021, 8, 22, 5, 3, 27, tzinfo=timezone.utc),
            id="isomillis-converter-no-state-with-start-start-is-start",
        ),
        pytest.param(
            IsoMillisConcurrentStreamStateConverter(),
            None,
            {"created_at": "2021-08-22T05:03:27.000Z"},
            datetime(2021, 8, 22, 5, 3, 27, tzinfo=timezone.utc),
            id="isomillis-converter-state-without-start-start-is-from-state",
        ),
        pytest.param(
            IsoMillisConcurrentStreamStateConverter(),
            "2022-08-22T05:03:27.000Z",
            {"created_at": "2021-08-22T05:03:27.000Z"},
            datetime(2022, 8, 22, 5, 3, 27, tzinfo=timezone.utc),
            id="isomillis-converter-state-before-start-start-is-start",
        ),
        pytest.param(
            IsoMillisConcurrentStreamStateConverter(),
            "2022-08-22T05:03:27.000Z",
            {"created_at": "2023-08-22T05:03:27.000Z"},
            datetime(2023, 8, 22, 5, 3, 27, tzinfo=timezone.utc),
            id="isomillis-converter-state-after-start-start-is-from-state",
        ),
    ],
)
def test_get_sync_start(converter, start, state, expected_start):
    assert converter._get_sync_start(CursorField("created_at"), state, start) == expected_start


@pytest.mark.parametrize(
    "converter, start, sequential_state, expected_output_state",
    [
        pytest.param(
            EpochValueConcurrentStreamStateConverter(),
            0,
            {},
            {
                "legacy": {},
                "slices": [
                    {
                        "start": EpochValueConcurrentStreamStateConverter().zero_value,
                        "end": EpochValueConcurrentStreamStateConverter().zero_value,
                    }
                ],
                "state_type": "date-range",
            },
            id="empty-input-state-epoch",
        ),
        pytest.param(
            EpochValueConcurrentStreamStateConverter(),
            1617030403,
            {"created": 1617030403},
            {
                "state_type": "date-range",
                "slices": [
                    {
                        "start": datetime(2021, 3, 29, 15, 6, 43, tzinfo=timezone.utc),
                        "end": datetime(2021, 3, 29, 15, 6, 43, tzinfo=timezone.utc),
                    }
                ],
                "legacy": {"created": 1617030403},
            },
            id="with-input-state-epoch",
        ),
        pytest.param(
            IsoMillisConcurrentStreamStateConverter(),
            "2020-01-01T00:00:00.000Z",
            {"created": "2021-08-22T05:03:27.000Z"},
            {
                "state_type": "date-range",
                "slices": [
                    {
                        "start": datetime(2021, 8, 22, 5, 3, 27, tzinfo=timezone.utc),
                        "end": datetime(2021, 8, 22, 5, 3, 27, tzinfo=timezone.utc),
                    }
                ],
                "legacy": {"created": "2021-08-22T05:03:27.000Z"},
            },
            id="with-input-state-isomillis",
        ),
    ],
)
def test_convert_from_sequential_state(converter, start, sequential_state, expected_output_state):
    comparison_format = "%Y-%m-%dT%H:%M:%S.%f"
    if expected_output_state["slices"]:
        _, conversion = converter.convert_from_sequential_state(CursorField("created"), sequential_state, start)
        assert conversion["state_type"] == expected_output_state["state_type"]
        assert conversion["legacy"] == expected_output_state["legacy"]
        for actual, expected in zip(conversion["slices"], expected_output_state["slices"]):
            assert actual["start"].strftime(comparison_format) == expected["start"].strftime(comparison_format)
            assert actual["end"].strftime(comparison_format) == expected["end"].strftime(comparison_format)
    else:
        _, conversion = converter.convert_from_sequential_state(CursorField("created"), sequential_state, start)
        assert conversion == expected_output_state


@pytest.mark.parametrize(
    "converter, concurrent_state, expected_output_state",
    [
        pytest.param(
            EpochValueConcurrentStreamStateConverter(),
            {
                "state_type": "date-range",
                "slices": [
                    {
                        "start": datetime(1970, 1, 3, 0, 0, 0, tzinfo=timezone.utc),
                        "end": datetime(2021, 3, 29, 15, 6, 43, tzinfo=timezone.utc),
                    }
                ],
            },
            {"created": 1617030403},
            id="epoch-single-slice",
        ),
        pytest.param(
            EpochValueConcurrentStreamStateConverter(),
            {
                "state_type": "date-range",
                "slices": [
                    {
                        "start": datetime(1970, 1, 1, 0, 0, 0, tzinfo=timezone.utc),
                        "end": datetime(2021, 3, 29, 15, 6, 43, tzinfo=timezone.utc),
                    },
                    {
                        "start": datetime(2020, 1, 1, 0, 0, 0, tzinfo=timezone.utc),
                        "end": datetime(2022, 3, 29, 15, 6, 43, tzinfo=timezone.utc),
                    },
                ],
            },
            {"created": 1648566403},
            id="epoch-overlapping-slices",
        ),
        pytest.param(
            EpochValueConcurrentStreamStateConverter(),
            {
                "state_type": "date-range",
                "slices": [
                    {
                        "start": datetime(1970, 1, 1, 0, 0, 0, tzinfo=timezone.utc),
                        "end": datetime(2021, 3, 29, 15, 6, 43, tzinfo=timezone.utc),
                    },
                    {
                        "start": datetime(2022, 1, 1, 0, 0, 0, tzinfo=timezone.utc),
                        "end": datetime(2023, 3, 29, 15, 6, 43, tzinfo=timezone.utc),
                    },
                ],
            },
            {"created": 1617030403},
            id="epoch-multiple-slices",
        ),
        pytest.param(
            IsoMillisConcurrentStreamStateConverter(),
            {
                "state_type": "date-range",
                "slices": [
                    {
                        "start": datetime(1970, 1, 3, 0, 0, 0, tzinfo=timezone.utc),
                        "end": datetime(2021, 3, 29, 15, 6, 43, tzinfo=timezone.utc),
                    }
                ],
            },
            {"created": "2021-03-29T15:06:43.000Z"},
            id="isomillis-single-slice",
        ),
        pytest.param(
            IsoMillisConcurrentStreamStateConverter(),
            {
                "state_type": "date-range",
                "slices": [
                    {
                        "start": datetime(1970, 1, 1, 0, 0, 0, tzinfo=timezone.utc),
                        "end": datetime(2021, 3, 29, 15, 6, 43, tzinfo=timezone.utc),
                    },
                    {
                        "start": datetime(2020, 1, 1, 0, 0, 0, tzinfo=timezone.utc),
                        "end": datetime(2022, 3, 29, 15, 6, 43, tzinfo=timezone.utc),
                    },
                ],
            },
            {"created": "2022-03-29T15:06:43.000Z"},
            id="isomillis-overlapping-slices",
        ),
        pytest.param(
            IsoMillisConcurrentStreamStateConverter(),
            {
                "state_type": "date-range",
                "slices": [
                    {
                        "start": datetime(1970, 1, 1, 0, 0, 0, tzinfo=timezone.utc),
                        "end": datetime(2021, 3, 29, 15, 6, 43, tzinfo=timezone.utc),
                    },
                    {
                        "start": datetime(2022, 1, 1, 0, 0, 0, tzinfo=timezone.utc),
                        "end": datetime(2023, 3, 29, 15, 6, 43, tzinfo=timezone.utc),
                    },
                ],
            },
            {"created": "2021-03-29T15:06:43.000Z"},
            id="isomillis-multiple-slices",
        ),
    ],
)
def test_convert_to_sequential_state(converter, concurrent_state, expected_output_state):
    assert converter.convert_to_sequential_state(CursorField("created"), concurrent_state) == expected_output_state


@pytest.mark.parametrize(
    "converter, concurrent_state, expected_output_state",
    [
        pytest.param(
            EpochValueConcurrentStreamStateConverter(),
            {
                "state_type": ConcurrencyCompatibleStateType.date_range.value,
                "start": EpochValueConcurrentStreamStateConverter().zero_value,
            },
            {"created": 0},
            id="empty-slices-epoch",
        ),
        pytest.param(
            IsoMillisConcurrentStreamStateConverter(),
            {
                "state_type": ConcurrencyCompatibleStateType.date_range.value,
                "start": datetime(2021, 8, 22, 5, 3, 27, tzinfo=timezone.utc),
            },
            {"created": "2021-08-22T05:03:27.000Z"},
            id="empty-slices-isomillis",
        ),
    ],
)
def test_convert_to_sequential_state_no_slices_returns_legacy_state(converter, concurrent_state, expected_output_state):
    with pytest.raises(RuntimeError):
        converter.convert_to_sequential_state(CursorField("created"), concurrent_state)



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

import logging
from typing import Any, Iterable, List, Mapping, Optional, Tuple, Union

from airbyte_cdk.models import SyncMode
from airbyte_cdk.sources import Source
from airbyte_cdk.sources.streams import Stream
from airbyte_cdk.sources.streams.availability_strategy import AvailabilityStrategy
from airbyte_cdk.sources.streams.core import StreamData

logger = logging.getLogger("airbyte")


class MockStream(Stream):
    def __init__(self, name: str) -> Stream:
        self._name = name

    @property
    def name(self) -> str:
        return self._name

    @property
    def primary_key(self) -> Optional[Union[str, List[str], List[List[str]]]]:
        pass

    def read_records(
        self,
        sync_mode: SyncMode,
        cursor_field: List[str] = None,
        stream_slice: Mapping[str, Any] = None,
        stream_state: Mapping[str, Any] = None,
    ) -> Iterable[StreamData]:
        pass


def test_no_availability_strategy():
    stream_1 = MockStream("stream")
    assert stream_1.availability_strategy is None

    stream_1_is_available, _ = stream_1.check_availability(logger)
    assert stream_1_is_available


def test_availability_strategy():
    class MockAvailabilityStrategy(AvailabilityStrategy):
        def check_availability(self, stream: Stream, logger: logging.Logger, source: Optional[Source]) -> Tuple[bool, any]:
            if stream.name == "available_stream":
                return True, None
            return False, f"Could not reach stream '{stream.name}'."

    class MockStreamWithAvailabilityStrategy(MockStream):
        @property
        def availability_strategy(self) -> Optional["AvailabilityStrategy"]:
            return MockAvailabilityStrategy()

    stream_1 = MockStreamWithAvailabilityStrategy("available_stream")
    stream_2 = MockStreamWithAvailabilityStrategy("unavailable_stream")

    for stream in [stream_1, stream_2]:
        assert isinstance(stream.availability_strategy, MockAvailabilityStrategy)

    stream_1_is_available, _ = stream_1.check_availability(logger)
    assert stream_1_is_available

    stream_2_is_available, reason = stream_2.check_availability(logger)
    assert not stream_2_is_available
    assert "Could not reach stream 'unavailable_stream'" in reason



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#


from typing import Any, List, Mapping, Optional, Set, Union


def combine_mappings(mappings: List[Optional[Union[Mapping[str, Any], str]]]) -> Union[Mapping[str, Any], str]:
    """
    Combine multiple mappings into a single mapping. If any of the mappings are a string, return
    that string. Raise errors in the following cases:
    * If there are duplicate keys across mappings
    * If there are multiple string mappings
    * If there are multiple mappings containing keys and one of them is a string
    """
    all_keys: List[Set[str]] = []
    for part in mappings:
        if part is None:
            continue
        keys = set(part.keys()) if not isinstance(part, str) else set()
        all_keys.append(keys)

    string_options = sum(isinstance(mapping, str) for mapping in mappings)
    # If more than one mapping is a string, raise a ValueError
    if string_options > 1:
        raise ValueError("Cannot combine multiple string options")

    if string_options == 1 and sum(len(keys) for keys in all_keys) > 0:
        raise ValueError("Cannot combine multiple options if one is a string")

    # If any mapping is a string, return it
    for mapping in mappings:
        if isinstance(mapping, str):
            return mapping

    # If there are duplicate keys across mappings, raise a ValueError
    intersection = set().union(*all_keys)
    if len(intersection) < sum(len(keys) for keys in all_keys):
        raise ValueError(f"Duplicate keys found: {intersection}")

    # Return the combined mappings
    return {key: value for mapping in mappings if mapping for key, value in mapping.items()}  # type: ignore # mapping can't be string here



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#


import inspect
import logging
import typing
from abc import ABC, abstractmethod
from functools import lru_cache
from typing import Any, Iterable, List, Mapping, MutableMapping, Optional, Tuple, Union

import airbyte_cdk.sources.utils.casing as casing
from airbyte_cdk.models import AirbyteMessage, AirbyteStream, ConfiguredAirbyteStream, SyncMode
from airbyte_cdk.models import Type as MessageType

# list of all possible HTTP methods which can be used for sending of request bodies
from airbyte_cdk.sources.utils.schema_helpers import InternalConfig, ResourceSchemaLoader
from airbyte_cdk.sources.utils.slice_logger import SliceLogger
from airbyte_cdk.sources.utils.transform import TransformConfig, TypeTransformer
from deprecated.classic import deprecated

if typing.TYPE_CHECKING:
    from airbyte_cdk.sources import Source
    from airbyte_cdk.sources.streams.availability_strategy import AvailabilityStrategy

# A stream's read method can return one of the following types:
# Mapping[str, Any]: The content of an AirbyteRecordMessage
# AirbyteMessage: An AirbyteMessage. Could be of any type
StreamData = Union[Mapping[str, Any], AirbyteMessage]

JsonSchema = Mapping[str, Any]

# Streams that only support full refresh don't have a suitable cursor so this sentinel
# value is used to indicate that stream should not load the incoming state value
FULL_REFRESH_SENTINEL_STATE_KEY = "__ab_full_refresh_state_message"


def package_name_from_class(cls: object) -> str:
    """Find the package name given a class name"""
    module = inspect.getmodule(cls)
    if module is not None:
        return module.__name__.split(".")[0]
    else:
        raise ValueError(f"Could not find package name for class {cls}")


class IncrementalMixin(ABC):
    """Mixin to make stream incremental.

    class IncrementalStream(Stream, IncrementalMixin):
        @property
        def state(self):
            return self._state

        @state.setter
        def state(self, value):
            self._state[self.cursor_field] = value[self.cursor_field]
    """

    @property
    @abstractmethod
    def state(self) -> MutableMapping[str, Any]:
        """State getter, should return state in form that can serialized to a string and send to the output
        as a STATE AirbyteMessage.

        A good example of a state is a cursor_value:
            {
                self.cursor_field: "cursor_value"
            }

         State should try to be as small as possible but at the same time descriptive enough to restore
         syncing process from the point where it stopped.
        """

    @state.setter
    @abstractmethod
    def state(self, value: MutableMapping[str, Any]) -> None:
        """State setter, accept state serialized by state getter."""


class Stream(ABC):
    """
    Base abstract class for an Airbyte Stream. Makes no assumption of the Stream's underlying transport protocol.
    """

    # Use self.logger in subclasses to log any messages
    @property
    def logger(self) -> logging.Logger:
        return logging.getLogger(f"airbyte.streams.{self.name}")

    # TypeTransformer object to perform output data transformation
    transformer: TypeTransformer = TypeTransformer(TransformConfig.NoTransform)

    @property
    def name(self) -> str:
        """
        :return: Stream name. By default this is the implementing class name, but it can be overridden as needed.
        """
        return casing.camel_to_snake(self.__class__.__name__)

    def get_error_display_message(self, exception: BaseException) -> Optional[str]:
        """
        Retrieves the user-friendly display message that corresponds to an exception.
        This will be called when encountering an exception while reading records from the stream, and used to build the AirbyteTraceMessage.

        The default implementation of this method does not return user-friendly messages for any exception type, but it should be overriden as needed.

        :param exception: The exception that was raised
        :return: A user-friendly message that indicates the cause of the error
        """
        return None

    def read(  # type: ignore  # ignoring typing for ConnectorStateManager because of circular dependencies
        self,
        configured_stream: ConfiguredAirbyteStream,
        logger: logging.Logger,
        slice_logger: SliceLogger,
        stream_state: MutableMapping[str, Any],
        state_manager,
        internal_config: InternalConfig,
    ) -> Iterable[StreamData]:
        sync_mode = configured_stream.sync_mode
        cursor_field = configured_stream.cursor_field

        slices = self.stream_slices(
            cursor_field=cursor_field,
            sync_mode=sync_mode,  # todo: change this interface to no longer rely on sync_mode for behavior
            stream_state=stream_state,
        )
        logger.debug(f"Processing stream slices for {self.name} (sync_mode: {sync_mode.name})", extra={"stream_slices": slices})

        has_slices = False
        record_counter = 0
        for _slice in slices:
            has_slices = True
            if slice_logger.should_log_slice_message(logger):
                yield slice_logger.create_slice_log_message(_slice)
            records = self.read_records(
                sync_mode=sync_mode,  # todo: change this interface to no longer rely on sync_mode for behavior
                stream_slice=_slice,
                stream_state=stream_state,
                cursor_field=cursor_field or None,
            )
            for record_data_or_message in records:
                yield record_data_or_message
                if isinstance(record_data_or_message, Mapping) or (
                    hasattr(record_data_or_message, "type") and record_data_or_message.type == MessageType.RECORD
                ):
                    record_data = record_data_or_message if isinstance(record_data_or_message, Mapping) else record_data_or_message.record
                    if self.cursor_field:
                        # Some connectors have streams that implement get_updated_state(), but do not define a cursor_field. This
                        # should be fixed on the stream implementation, but we should also protect against this in the CDK as well
                        stream_state = self.get_updated_state(stream_state, record_data)
                    record_counter += 1

                    if sync_mode == SyncMode.incremental:
                        # Checkpoint intervals are a bit controversial, but see below comment about why we're gating it right now
                        checkpoint_interval = self.state_checkpoint_interval
                        if checkpoint_interval and record_counter % checkpoint_interval == 0:
                            airbyte_state_message = self._checkpoint_state(stream_state, state_manager)
                            yield airbyte_state_message

                    if internal_config.is_limit_reached(record_counter):
                        break

            if sync_mode == SyncMode.incremental:
                # Even though right now, only incremental streams running as incremental mode will emit periodic checkpoints. Rather than
                # overhaul how refresh interacts with the platform, this positions the code so that once we want to start emitting
                # periodic checkpoints in full refresh mode it can be done here
                airbyte_state_message = self._checkpoint_state(stream_state, state_manager)
                yield airbyte_state_message

        if not has_slices or sync_mode == SyncMode.full_refresh:
            if sync_mode == SyncMode.full_refresh:
                # We use a dummy state if there is no suitable value provided by full_refresh streams that do not have a valid cursor.
                # Incremental streams running full_refresh mode emit a meaningful state
                stream_state = stream_state or {FULL_REFRESH_SENTINEL_STATE_KEY: True}

            # We should always emit a final state message for full refresh sync or streams that do not have any slices
            airbyte_state_message = self._checkpoint_state(stream_state, state_manager)
            yield airbyte_state_message

    @abstractmethod
    def read_records(
        self,
        sync_mode: SyncMode,
        cursor_field: Optional[List[str]] = None,
        stream_slice: Optional[Mapping[str, Any]] = None,
        stream_state: Optional[Mapping[str, Any]] = None,
    ) -> Iterable[StreamData]:
        """
        This method should be overridden by subclasses to read records based on the inputs
        """

    @lru_cache(maxsize=None)
    def get_json_schema(self) -> Mapping[str, Any]:
        """
        :return: A dict of the JSON schema representing this stream.

        The default implementation of this method looks for a JSONSchema file with the same name as this stream's "name" property.
        Override as needed.
        """
        # TODO show an example of using pydantic to define the JSON schema, or reading an OpenAPI spec
        return ResourceSchemaLoader(package_name_from_class(self.__class__)).get_schema(self.name)

    def as_airbyte_stream(self) -> AirbyteStream:
        stream = AirbyteStream(name=self.name, json_schema=dict(self.get_json_schema()), supported_sync_modes=[SyncMode.full_refresh])

        if self.namespace:
            stream.namespace = self.namespace

        if self.supports_incremental:
            stream.source_defined_cursor = self.source_defined_cursor
            stream.supported_sync_modes.append(SyncMode.incremental)  # type: ignore
            stream.default_cursor_field = self._wrapped_cursor_field()

        keys = Stream._wrapped_primary_key(self.primary_key)
        if keys and len(keys) > 0:
            stream.source_defined_primary_key = keys

        return stream

    @property
    def supports_incremental(self) -> bool:
        """
        :return: True if this stream supports incrementally reading data
        """
        return len(self._wrapped_cursor_field()) > 0

    def _wrapped_cursor_field(self) -> List[str]:
        return [self.cursor_field] if isinstance(self.cursor_field, str) else self.cursor_field

    @property
    def cursor_field(self) -> Union[str, List[str]]:
        """
        Override to return the default cursor field used by this stream e.g: an API entity might always use created_at as the cursor field.
        :return: The name of the field used as a cursor. If the cursor is nested, return an array consisting of the path to the cursor.
        """
        return []

    @property
    def namespace(self) -> Optional[str]:
        """
        Override to return the namespace of this stream, e.g. the Postgres schema which this stream will emit records for.
        :return: A string containing the name of the namespace.
        """
        return None

    @property
    def source_defined_cursor(self) -> bool:
        """
        Return False if the cursor can be configured by the user.
        """
        return True

    def check_availability(self, logger: logging.Logger, source: Optional["Source"] = None) -> Tuple[bool, Optional[str]]:
        """
        Checks whether this stream is available.

        :param logger: source logger
        :param source: (optional) source
        :return: A tuple of (boolean, str). If boolean is true, then this stream
          is available, and no str is required. Otherwise, this stream is unavailable
          for some reason and the str should describe what went wrong and how to
          resolve the unavailability, if possible.
        """
        if self.availability_strategy:
            return self.availability_strategy.check_availability(self, logger, source)
        return True, None

    @property
    def availability_strategy(self) -> Optional["AvailabilityStrategy"]:
        """
        :return: The AvailabilityStrategy used to check whether this stream is available.
        """
        return None

    @property
    @abstractmethod
    def primary_key(self) -> Optional[Union[str, List[str], List[List[str]]]]:
        """
        :return: string if single primary key, list of strings if composite primary key, list of list of strings if composite primary key consisting of nested fields.
          If the stream has no primary keys, return None.
        """

    def stream_slices(
        self, *, sync_mode: SyncMode, cursor_field: Optional[List[str]] = None, stream_state: Optional[Mapping[str, Any]] = None
    ) -> Iterable[Optional[Mapping[str, Any]]]:
        """
        Override to define the slices for this stream. See the stream slicing section of the docs for more information.

        :param sync_mode:
        :param cursor_field:
        :param stream_state:
        :return:
        """
        return [None]

    @property
    def state_checkpoint_interval(self) -> Optional[int]:
        """
        Decides how often to checkpoint state (i.e: emit a STATE message). E.g: if this returns a value of 100, then state is persisted after reading
        100 records, then 200, 300, etc.. A good default value is 1000 although your mileage may vary depending on the underlying data source.

        Checkpointing a stream avoids re-reading records in the case a sync is failed or cancelled.

        return None if state should not be checkpointed e.g: because records returned from the underlying data source are not returned in
        ascending order with respect to the cursor field. This can happen if the source does not support reading records in ascending order of
        created_at date (or whatever the cursor is). In those cases, state must only be saved once the full stream has been read.
        """
        return None

    @deprecated(version="0.1.49", reason="You should use explicit state property instead, see IncrementalMixin docs.")
    def get_updated_state(
        self, current_stream_state: MutableMapping[str, Any], latest_record: Mapping[str, Any]
    ) -> MutableMapping[str, Any]:
        """Override to extract state from the latest record. Needed to implement incremental sync.

        Inspects the latest record extracted from the data source and the current state object and return an updated state object.

        For example: if the state object is based on created_at timestamp, and the current state is {'created_at': 10}, and the latest_record is
        {'name': 'octavia', 'created_at': 20 } then this method would return {'created_at': 20} to indicate state should be updated to this object.

        :param current_stream_state: The stream's current state object
        :param latest_record: The latest record extracted from the stream
        :return: An updated state object
        """
        return {}

    def log_stream_sync_configuration(self) -> None:
        """
        Logs the configuration of this stream.
        """
        self.logger.debug(
            f"Syncing stream instance: {self.name}",
            extra={
                "primary_key": self.primary_key,
                "cursor_field": self.cursor_field,
            },
        )

    @staticmethod
    def _wrapped_primary_key(keys: Optional[Union[str, List[str], List[List[str]]]]) -> Optional[List[List[str]]]:
        """
        :return: wrap the primary_key property in a list of list of strings required by the Airbyte Stream object.
        """
        if not keys:
            return None

        if isinstance(keys, str):
            return [[keys]]
        elif isinstance(keys, list):
            wrapped_keys = []
            for component in keys:
                if isinstance(component, str):
                    wrapped_keys.append([component])
                elif isinstance(component, list):
                    wrapped_keys.append(component)
                else:
                    raise ValueError(f"Element must be either list or str. Got: {type(component)}")
            return wrapped_keys
        else:
            raise ValueError(f"Element must be either list or str. Got: {type(keys)}")

    def _checkpoint_state(  # type: ignore  # ignoring typing for ConnectorStateManager because of circular dependencies
        self,
        stream_state: Mapping[str, Any],
        state_manager,
    ) -> AirbyteMessage:
        # First attempt to retrieve the current state using the stream's state property. We receive an AttributeError if the state
        # property is not implemented by the stream instance and as a fallback, use the stream_state retrieved from the stream
        # instance's deprecated get_updated_state() method.
        try:
            state_manager.update_state_for_stream(
                self.name, self.namespace, self.state  # type: ignore # we know the field might not exist...
            )

        except AttributeError:
            state_manager.update_state_for_stream(self.name, self.namespace, stream_state)
        return state_manager.create_state_message(self.name, self.namespace)



# Copyright (c) 2023 Airbyte, Inc., all rights reserved.

from unittest import TestCase
from unittest.mock import Mock

from airbyte_cdk.test.mock_http.matcher import HttpRequestMatcher
from airbyte_cdk.test.mock_http.request import HttpRequest


class HttpRequestMatcherTest(TestCase):
    def setUp(self) -> None:
        self._a_request = Mock(spec=HttpRequest)
        self._another_request = Mock(spec=HttpRequest)
        self._request_to_match = Mock(spec=HttpRequest)
        self._matcher = HttpRequestMatcher(self._request_to_match, 1)

    def test_given_request_matches_when_matches_then_has_expected_match_count(self):
        self._a_request.matches.return_value = True
        self._matcher.matches(self._a_request)
        assert self._matcher.has_expected_match_count()

    def test_given_request_does_not_match_when_matches_then_does_not_have_expected_match_count(self):
        self._a_request.matches.return_value = False
        self._matcher.matches(self._a_request)

        assert not self._matcher.has_expected_match_count()
        assert self._matcher.actual_number_of_matches == 0

    def test_given_many_requests_with_some_match_when_matches_then_has_expected_match_count(self):
        self._a_request.matches.return_value = True
        self._another_request.matches.return_value = False
        self._matcher.matches(self._a_request)
        self._matcher.matches(self._another_request)

        assert self._matcher.has_expected_match_count()
        assert self._matcher.actual_number_of_matches == 1

    def test_given_expected_number_of_requests_met_when_matches_then_has_expected_match_count(self):
        _matcher = HttpRequestMatcher(self._request_to_match, 2)
        self._a_request.matches.return_value = True
        _matcher.matches(self._a_request)
        _matcher.matches(self._a_request)

        assert _matcher.has_expected_match_count()
        assert _matcher.actual_number_of_matches == 2

    def test_given_expected_number_of_requests_not_met_when_matches_then_does_not_have_expected_match_count(self):
        _matcher = HttpRequestMatcher(self._request_to_match, 2)
        self._a_request.matches.side_effect = [True, False]
        _matcher.matches(self._a_request)
        _matcher.matches(self._a_request)

        assert not _matcher.has_expected_match_count()



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

import traceback
import unittest
from datetime import datetime, timezone
from typing import Any, Iterable, Iterator, Mapping
from unittest.mock import Mock

import pytest
from airbyte_cdk.models import AirbyteLogMessage, AirbyteMessage, Level
from airbyte_cdk.models import Type as MessageType
from airbyte_cdk.sources.file_based.availability_strategy import AbstractFileBasedAvailabilityStrategy
from airbyte_cdk.sources.file_based.discovery_policy import AbstractDiscoveryPolicy
from airbyte_cdk.sources.file_based.exceptions import FileBasedErrorsCollector, FileBasedSourceError
from airbyte_cdk.sources.file_based.file_based_stream_reader import AbstractFileBasedStreamReader
from airbyte_cdk.sources.file_based.file_types.file_type_parser import FileTypeParser
from airbyte_cdk.sources.file_based.remote_file import RemoteFile
from airbyte_cdk.sources.file_based.schema_validation_policies import AbstractSchemaValidationPolicy
from airbyte_cdk.sources.file_based.stream.cursor import AbstractFileBasedCursor
from airbyte_cdk.sources.file_based.stream.default_file_based_stream import DefaultFileBasedStream
from airbyte_cdk.utils.traced_exception import AirbyteTracedException


class MockFormat:
    pass


@pytest.mark.parametrize(
    "input_schema, expected_output",
    [
        pytest.param({}, {}, id="empty-schema"),
        pytest.param(
            {"type": "string"},
            {"type": ["null", "string"]},
            id="simple-schema",
        ),
        pytest.param(
            {"type": ["string"]},
            {"type": ["null", "string"]},
            id="simple-schema-list-type",
        ),
        pytest.param(
            {"type": ["null", "string"]},
            {"type": ["null", "string"]},
            id="simple-schema-already-has-null",
        ),
        pytest.param(
            {"properties": {"type": "string"}},
            {"properties": {"type": ["null", "string"]}},
            id="nested-schema",
        ),
        pytest.param(
            {"items": {"type": "string"}},
            {"items": {"type": ["null", "string"]}},
            id="array-schema",
        ),
        pytest.param(
            {"type": "object", "properties": {"prop": {"type": "string"}}},
            {
                "type": ["null", "object"],
                "properties": {"prop": {"type": ["null", "string"]}},
            },
            id="deeply-nested-schema",
        ),
    ],
)
def test_fill_nulls(input_schema: Mapping[str, Any], expected_output: Mapping[str, Any]) -> None:
    assert DefaultFileBasedStream._fill_nulls(input_schema) == expected_output


class DefaultFileBasedStreamTest(unittest.TestCase):
    _NOW = datetime(2022, 10, 22, tzinfo=timezone.utc)
    _A_RECORD = {"a_record": 1}

    def setUp(self) -> None:
        self._stream_config = Mock()
        self._stream_config.format = MockFormat()
        self._stream_config.name = "a stream name"
        self._catalog_schema = Mock()
        self._stream_reader = Mock(spec=AbstractFileBasedStreamReader)
        self._availability_strategy = Mock(spec=AbstractFileBasedAvailabilityStrategy)
        self._discovery_policy = Mock(spec=AbstractDiscoveryPolicy)
        self._parser = Mock(spec=FileTypeParser)
        self._validation_policy = Mock(spec=AbstractSchemaValidationPolicy)
        self._validation_policy.name = "validation policy name"
        self._cursor = Mock(spec=AbstractFileBasedCursor)

        self._stream = DefaultFileBasedStream(
            config=self._stream_config,
            catalog_schema=self._catalog_schema,
            stream_reader=self._stream_reader,
            availability_strategy=self._availability_strategy,
            discovery_policy=self._discovery_policy,
            parsers={MockFormat: self._parser},
            validation_policy=self._validation_policy,
            cursor=self._cursor,
            errors_collector=FileBasedErrorsCollector(),
        )

    def test_when_read_records_from_slice_then_return_records(self) -> None:
        self._parser.parse_records.return_value = [self._A_RECORD]
        messages = list(self._stream.read_records_from_slice({"files": [RemoteFile(uri="uri", last_modified=self._NOW)]}))
        assert list(map(lambda message: message.record.data["data"], messages)) == [self._A_RECORD]

    def test_given_exception_when_read_records_from_slice_then_do_process_other_files(
        self,
    ) -> None:
        """
        The current behavior for source-s3 v3 does not fail sync on some errors and hence, we will keep this behaviour for now. One example
        we can easily reproduce this is by having a file with gzip extension that is not actually a gzip file. The reader will fail to open
        the file but the sync won't fail.
        Ticket: https://github.com/airbytehq/airbyte/issues/29680
        """
        self._parser.parse_records.side_effect = [
            ValueError("An error"),
            [self._A_RECORD],
        ]

        messages = list(
            self._stream.read_records_from_slice(
                {
                    "files": [
                        RemoteFile(uri="invalid_file", last_modified=self._NOW),
                        RemoteFile(uri="valid_file", last_modified=self._NOW),
                    ]
                }
            )
        )

        assert messages[0].log.level == Level.ERROR
        assert messages[1].record.data["data"] == self._A_RECORD

    def test_given_traced_exception_when_read_records_from_slice_then_fail(
        self,
    ) -> None:
        """
        When a traced exception is raised, the stream shouldn't try to handle but pass it on to the caller.
        """
        self._parser.parse_records.side_effect = [AirbyteTracedException("An error")]

        with pytest.raises(AirbyteTracedException):
            list(
                self._stream.read_records_from_slice(
                    {
                        "files": [
                            RemoteFile(uri="invalid_file", last_modified=self._NOW),
                            RemoteFile(uri="valid_file", last_modified=self._NOW),
                        ]
                    }
                )
            )

    def test_given_exception_after_skipping_records_when_read_records_from_slice_then_send_warning(
        self,
    ) -> None:
        self._stream_config.schemaless = False
        self._validation_policy.record_passes_validation_policy.return_value = False
        self._parser.parse_records.side_effect = [self._iter([self._A_RECORD, ValueError("An error")])]

        messages = list(
            self._stream.read_records_from_slice(
                {
                    "files": [
                        RemoteFile(uri="invalid_file", last_modified=self._NOW),
                        RemoteFile(uri="valid_file", last_modified=self._NOW),
                    ]
                }
            )
        )

        assert messages[0].log.level == Level.ERROR
        assert messages[1].log.level == Level.WARN

    def test_override_max_n_files_for_schema_inference_is_respected(self) -> None:
        self._discovery_policy.n_concurrent_requests = 1
        self._discovery_policy.get_max_n_files_for_schema_inference.return_value = 3
        self._stream.config.input_schema = None
        self._stream.config.schemaless = None
        self._parser.infer_schema.return_value = {"data": {"type": "string"}}
        files = [RemoteFile(uri=f"file{i}", last_modified=self._NOW) for i in range(10)]
        self._stream_reader.get_matching_files.return_value = files

        schema = self._stream.get_json_schema()

        assert schema == {
            "type": "object",
            "properties": {
                "_ab_source_file_last_modified": {"type": "string"},
                "_ab_source_file_url": {"type": "string"},
                "data": {"type": ["null", "string"]},
            },
        }
        assert self._parser.infer_schema.call_count == 3

    def _iter(self, x: Iterable[Any]) -> Iterator[Any]:
        for item in x:
            if isinstance(item, Exception):
                raise item
            yield item


class TestFileBasedErrorCollector:
    test_error_collector: FileBasedErrorsCollector = FileBasedErrorsCollector()

    @pytest.mark.parametrize(
        "stream, file, line_no, n_skipped, collector_expected_len",
        (
            ("stream_1", "test.csv", 1, 1, 1),
            ("stream_2", "test2.csv", 2, 2, 2),
        ),
        ids=[
            "Single error",
            "Multiple errors",
        ],
    )
    def test_collect_parsing_error(self, stream, file, line_no, n_skipped, collector_expected_len) -> None:
        test_error_pattern = "Error parsing record."
        # format the error body
        test_error = (
            AirbyteMessage(
                type=MessageType.LOG,
                log=AirbyteLogMessage(
                    level=Level.ERROR,
                    message=f"{FileBasedSourceError.ERROR_PARSING_RECORD.value} stream={stream} file={file} line_no={line_no} n_skipped={n_skipped}",
                    stack_trace=traceback.format_exc(),
                ),
            ),
        )
        # collecting the error
        self.test_error_collector.collect(test_error)
        # check the error has been collected
        assert len(self.test_error_collector.errors) == collector_expected_len
        # check for the patern presence for the collected errors
        for error in self.test_error_collector.errors:
            assert test_error_pattern in error[0].log.message

    def test_yield_and_raise_collected(self) -> None:
        # we expect the following method will raise the AirbyteTracedException
        with pytest.raises(AirbyteTracedException) as parse_error:
            list(self.test_error_collector.yield_and_raise_collected())
        assert parse_error.value.message == "Some errors occured while reading from the source."
        assert parse_error.value.internal_message == "Please check the logged errors for more information."



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

import logging
from datetime import datetime, timedelta
from typing import Any, Iterable, MutableMapping, Optional

from airbyte_cdk.sources.file_based.config.file_based_stream_config import FileBasedStreamConfig
from airbyte_cdk.sources.file_based.remote_file import RemoteFile
from airbyte_cdk.sources.file_based.stream.cursor.abstract_file_based_cursor import AbstractFileBasedCursor
from airbyte_cdk.sources.file_based.types import StreamState


class DefaultFileBasedCursor(AbstractFileBasedCursor):
    DEFAULT_DAYS_TO_SYNC_IF_HISTORY_IS_FULL = 3
    DEFAULT_MAX_HISTORY_SIZE = 10_000
    DATE_TIME_FORMAT = "%Y-%m-%dT%H:%M:%S.%fZ"
    CURSOR_FIELD = "_ab_source_file_last_modified"

    def __init__(self, stream_config: FileBasedStreamConfig, **_: Any):
        super().__init__(stream_config)
        self._file_to_datetime_history: MutableMapping[str, str] = {}
        self._time_window_if_history_is_full = timedelta(
            days=stream_config.days_to_sync_if_history_is_full or self.DEFAULT_DAYS_TO_SYNC_IF_HISTORY_IS_FULL
        )

        if self._time_window_if_history_is_full <= timedelta():
            raise ValueError(f"days_to_sync_if_history_is_full must be a positive timedelta, got {self._time_window_if_history_is_full}")

        self._start_time = self._compute_start_time()
        self._initial_earliest_file_in_history: Optional[RemoteFile] = None

    def set_initial_state(self, value: StreamState) -> None:
        self._file_to_datetime_history = value.get("history", {})
        self._start_time = self._compute_start_time()
        self._initial_earliest_file_in_history = self._compute_earliest_file_in_history()

    def add_file(self, file: RemoteFile) -> None:
        self._file_to_datetime_history[file.uri] = file.last_modified.strftime(self.DATE_TIME_FORMAT)
        if len(self._file_to_datetime_history) > self.DEFAULT_MAX_HISTORY_SIZE:
            # Get the earliest file based on its last modified date and its uri
            oldest_file = self._compute_earliest_file_in_history()
            if oldest_file:
                del self._file_to_datetime_history[oldest_file.uri]
            else:
                raise Exception(
                    "The history is full but there is no files in the history. This should never happen and might be indicative of a bug in the CDK."
                )

    def get_state(self) -> StreamState:
        state = {"history": self._file_to_datetime_history, self.CURSOR_FIELD: self._get_cursor()}
        return state

    def _get_cursor(self) -> Optional[str]:
        """
        Returns the cursor value.

        Files are synced in order of last-modified with secondary sort on filename, so the cursor value is
        a string joining the last-modified timestamp of the last synced file and the name of the file.
        """
        if self._file_to_datetime_history.items():
            filename, timestamp = max(self._file_to_datetime_history.items(), key=lambda x: (x[1], x[0]))
            return f"{timestamp}_{filename}"
        return None

    def _is_history_full(self) -> bool:
        """
        Returns true if the state's history is full, meaning new entries will start to replace old entries.
        """
        return len(self._file_to_datetime_history) >= self.DEFAULT_MAX_HISTORY_SIZE

    def _should_sync_file(self, file: RemoteFile, logger: logging.Logger) -> bool:
        if file.uri in self._file_to_datetime_history:
            # If the file's uri is in the history, we should sync the file if it has been modified since it was synced
            updated_at_from_history = datetime.strptime(self._file_to_datetime_history[file.uri], self.DATE_TIME_FORMAT)
            if file.last_modified < updated_at_from_history:
                logger.warning(
                    f"The file {file.uri}'s last modified date is older than the last time it was synced. This is unexpected. Skipping the file."
                )
            else:
                return file.last_modified > updated_at_from_history
            return file.last_modified > updated_at_from_history
        if self._is_history_full():
            if self._initial_earliest_file_in_history is None:
                return True
            if file.last_modified > self._initial_earliest_file_in_history.last_modified:
                # If the history is partial and the file's datetime is strictly greater than the earliest file in the history,
                # we should sync it
                return True
            elif file.last_modified == self._initial_earliest_file_in_history.last_modified:
                # If the history is partial and the file's datetime is equal to the earliest file in the history,
                # we should sync it if its uri is strictly greater than the earliest file in the history
                return file.uri > self._initial_earliest_file_in_history.uri
            else:
                # Otherwise, only sync the file if it has been modified since the start of the time window
                return file.last_modified >= self.get_start_time()
        else:
            # The file is not in the history and the history is complete. We know we need to sync the file
            return True

    def get_files_to_sync(self, all_files: Iterable[RemoteFile], logger: logging.Logger) -> Iterable[RemoteFile]:
        if self._is_history_full():
            logger.warning(
                f"The state history is full. "
                f"This sync and future syncs won't be able to use the history to filter out duplicate files. "
                f"It will instead use the time window of {self._time_window_if_history_is_full} to filter out files."
            )
        for f in all_files:
            if self._should_sync_file(f, logger):
                yield f

    def get_start_time(self) -> datetime:
        return self._start_time

    def _compute_earliest_file_in_history(self) -> Optional[RemoteFile]:
        if self._file_to_datetime_history:
            filename, last_modified = min(self._file_to_datetime_history.items(), key=lambda f: (f[1], f[0]))
            return RemoteFile(uri=filename, last_modified=datetime.strptime(last_modified, self.DATE_TIME_FORMAT))
        else:
            return None

    def _compute_start_time(self) -> datetime:
        if not self._file_to_datetime_history:
            return datetime.min
        else:
            earliest = min(self._file_to_datetime_history.values())
            earliest_dt = datetime.strptime(earliest, self.DATE_TIME_FORMAT)
            if self._is_history_full():
                time_window = datetime.now() - self._time_window_if_history_is_full
                earliest_dt = min(earliest_dt, time_window)
            return earliest_dt



#
# Copyright (c) 2021 Airbyte, Inc., all rights reserved.
#

import dpath.options

from .abstract_source import AbstractSource
from .config import BaseConfig
from .source import Source

# As part of the CDK sources, we do not control what the APIs return and it is possible that a key is empty.
# Reasons why we are doing this at the airbyte_cdk level:
# * As of today, all the use cases should allow for empty keys
#     * Cases as of 2023-08-31: oauth/session token provider responses, extractor, transformation and substream)
# * The behavior is explicit at the package level and not hidden in every package that needs dpath.options.ALLOW_EMPTY_STRING_KEYS = True
# There is a downside in enforcing this option preemptively in the module __init__.py: the runtime code will import dpath even though the it
# might not need dpath leading to longer initialization time.
# There is a downside in using dpath as a library since the options are global: if we have two pieces of code that want different options,
# this will not be thread-safe.
dpath.options.ALLOW_EMPTY_STRING_KEYS = True

__all__ = ["AbstractSource", "BaseConfig", "Source"]



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#


from airbyte_cdk.utils.traced_exception import AirbyteTracedException
from unit_tests.sources.file_based.scenarios.file_based_source_builder import FileBasedSourceBuilder
from unit_tests.sources.file_based.scenarios.scenario_builder import TestScenarioBuilder

_base_single_stream_scenario = (
    TestScenarioBuilder()
    .set_source_builder(
        FileBasedSourceBuilder()
        .set_files(
            {
                "a.csv": {
                    "contents": [
                        ("col1", "col2"),
                        ("val_a_11", "1"),
                        ("val_a_12", "2"),
                    ],
                    "last_modified": "2023-06-05T03:54:07.000Z",
                },
                "b.csv": {  # The records in this file do not conform to the schema
                    "contents": [
                        ("col1", "col2"),
                        ("val_b_11", "this is text that will trigger validation policy"),
                        ("val_b_12", "2"),
                    ],
                    "last_modified": "2023-06-05T03:54:07.000Z",
                },
                "c.csv": {
                    "contents": [
                        ("col1",),
                        ("val_c_11",),
                        ("val_c_12", "val_c_22"),  # This record is not parsable
                        ("val_c_13",),
                    ],
                    "last_modified": "2023-06-05T03:54:07.000Z",
                },
                "d.csv": {
                    "contents": [
                        ("col1",),
                        ("val_d_11",),
                    ],
                    "last_modified": "2023-06-05T03:54:07.000Z",
                },
            }
        )
        .set_file_type("csv")
    )
    .set_expected_catalog(
        {
            "streams": [
                {
                    "default_cursor_field": ["_ab_source_file_last_modified"],
                    "json_schema": {
                        "type": "object",
                        "properties": {
                            "col1": {
                                "type": "string",
                            },
                            "col2": {
                                "type": "integer",
                            },
                            "_ab_source_file_last_modified": {"type": "string"},
                            "_ab_source_file_url": {"type": "string"},
                        },
                    },
                    "name": "stream1",
                    "source_defined_cursor": True,
                    "supported_sync_modes": ["full_refresh", "incremental"],
                }
            ]
        }
    )
)


_base_multi_stream_scenario = (
    TestScenarioBuilder()
    .set_source_builder(
        FileBasedSourceBuilder()
        .set_files(
            {
                "a/a1.csv": {
                    "contents": [
                        ("col1", "col2"),
                        ("val_aa1_11", "1"),
                        ("val_aa1_12", "2"),
                    ],
                    "last_modified": "2023-06-05T03:54:07.000Z",
                },
                "a/a2.csv": {
                    "contents": [
                        ("col1", "col2"),
                        ("val_aa2_11", "this is text that will trigger validation policy"),
                        ("val_aa2_12", "2"),
                    ],
                    "last_modified": "2023-06-05T03:54:07.000Z",
                },
                "a/a3.csv": {
                    "contents": [
                        ("col1",),
                        ("val_aa3_11",),
                        ("val_aa3_12", "val_aa3_22"),  # This record is not parsable
                        ("val_aa3_13",),
                    ],
                    "last_modified": "2023-06-05T03:54:07.000Z",
                },
                "a/a4.csv": {
                    "contents": [
                        ("col1",),
                        ("val_aa4_11",),
                    ],
                    "last_modified": "2023-06-05T03:54:07.000Z",
                },
                "b/b1.csv": {  # The records in this file do not conform to the schema
                    "contents": [
                        ("col1", "col2"),
                        ("val_bb1_11", "1"),
                        ("val_bb1_12", "2"),
                    ],
                    "last_modified": "2023-06-05T03:54:07.000Z",
                },
                "b/b2.csv": {
                    "contents": [
                        ("col1", "col2"),
                        ("val_bb2_11", "this is text that will trigger validation policy"),
                        ("val_bb2_12", "2"),
                    ],
                    "last_modified": "2023-06-05T03:54:07.000Z",
                },
                "b/b3.csv": {
                    "contents": [
                        ("col1",),
                        ("val_bb3_11",),
                        ("val_bb3_12",),
                    ],
                    "last_modified": "2023-06-05T03:54:07.000Z",
                },
            }
        )
        .set_file_type("csv")
    )
    .set_expected_catalog(
        {
            "streams": [
                {
                    "default_cursor_field": ["_ab_source_file_last_modified"],
                    "json_schema": {
                        "type": "object",
                        "properties": {
                            "col1": {
                                "type": "string",
                            },
                            "col2": {
                                "type": "integer",
                            },
                            "_ab_source_file_last_modified": {"type": "string"},
                            "_ab_source_file_url": {"type": "string"},
                        },
                    },
                    "name": "stream1",
                    "source_defined_cursor": True,
                    "supported_sync_modes": ["full_refresh", "incremental"],
                },
                {
                    "json_schema": {
                        "default_cursor_field": ["_ab_source_file_last_modified"],
                        "type": "object",
                        "properties": {
                            "col1": {
                                "type": "string",
                            },
                            "col2": {
                                "type": "integer",
                            },
                            "_ab_source_file_last_modified": {"type": "string"},
                            "_ab_source_file_url": {"type": "string"},
                        },
                    },
                    "name": "stream2",
                    "source_defined_cursor": True,
                    "supported_sync_modes": ["full_refresh", "incremental"],
                },
            ]
        }
    )
)


skip_record_scenario_single_stream = (
    _base_single_stream_scenario.copy()
    .set_name("skip_record_scenario_single_stream")
    .set_config(
        {
            "streams": [
                {
                    "name": "stream1",
                    "format": {"filetype": "csv"},
                    "globs": ["*.csv"],
                    "validation_policy": "Skip Record",
                }
            ]
        }
    )
    .set_expected_records(
        [
            {
                "data": {
                    "col1": "val_a_11",
                    "col2": 1,
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "a.csv",
                },
                "stream": "stream1",
            },
            {
                "data": {
                    "col1": "val_a_12",
                    "col2": 2,
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "a.csv",
                },
                "stream": "stream1",
            },
            # {"data": {"col1": "val_b_11", "col2": "this is text that will trigger validation policy", "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z", "_ab_source_file_url": "b.csv"}, "stream": "stream1"},  # This record is skipped because it does not conform
            {
                "data": {
                    "col1": "val_b_12",
                    "col2": 2,
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "b.csv",
                },
                "stream": "stream1",
            },
            {
                "data": {
                    "col1": "val_c_11",
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "c.csv",
                },
                "stream": "stream1",
            },
            # {"data": {"col1": "val_c_12", None: "val_c_22", "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z", "_ab_source_file_url": "c.csv"}, "stream": "stream1"},  # This record is malformed so should not be emitted
            # {"data": {"col1": "val_c_13", "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z", "_ab_source_file_url": "c.csv"}, "stream": "stream1"},  # Skipped since previous record is malformed
            {
                "data": {
                    "col1": "val_d_11",
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "d.csv",
                },
                "stream": "stream1",
            },
        ]
    )
    .set_expected_logs(
        {
            "read": [
                {
                    "level": "WARN",
                    "message": "Records in file did not pass validation policy. stream=stream1 file=b.csv n_skipped=1 validation_policy=skip_record",
                },
                {
                    "level": "ERROR",
                    "message": "Error parsing record. This could be due to a mismatch between the config's file type and the actual file type, or because the file or record is not parseable. stream=stream1 file=c.csv line_no=2 n_skipped=0",
                },
                {
                    "level": "WARN",
                    "message": "Could not cast the value to the expected type.: col2: value=this is text that will trigger validation policy,expected_type=integer",
                },
            ]
        }
    )
    .set_expected_read_error(
        AirbyteTracedException,
        "Please check the logged errors for more information.",
    )
).build()


skip_record_scenario_multi_stream = (
    _base_multi_stream_scenario.copy()
    .set_name("skip_record_scenario_multi_stream")
    .set_config(
        {
            "streams": [
                {
                    "name": "stream1",
                    "format": {"filetype": "csv"},
                    "globs": ["a/*.csv"],
                    "validation_policy": "Skip Record",
                },
                {
                    "name": "stream2",
                    "format": {"filetype": "csv"},
                    "globs": ["b/*.csv"],
                    "validation_policy": "Skip Record",
                },
            ]
        }
    )
    .set_expected_records(
        [
            {
                "data": {
                    "col1": "val_aa1_11",
                    "col2": 1,
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "a/a1.csv",
                },
                "stream": "stream1",
            },
            {
                "data": {
                    "col1": "val_aa1_12",
                    "col2": 2,
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "a/a1.csv",
                },
                "stream": "stream1",
            },
            # {"data": {"col1": "val_aa2_11", "col2": "this is text that will trigger validation policy", "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z", "_ab_source_file_url": "a/a2.csv"}, "stream": "stream1"},  # This record is skipped because it does not conform
            {
                "data": {
                    "col1": "val_aa2_12",
                    "col2": 2,
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "a/a2.csv",
                },
                "stream": "stream1",
            },
            {
                "data": {
                    "col1": "val_aa3_11",
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "a/a3.csv",
                },
                "stream": "stream1",
            },
            # {"data": {"col1": "val_aa3_12", None: "val_aa3_22", "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z", "_ab_source_file_url": "a/a3.csv"}, "stream": "stream1"},  # This record is malformed so should not be emitted
            # {"data": {"col1": "val_aa3_13", "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z", "_ab_source_file_url": "a/a3.csv"}, "stream": "stream1"},  # Skipped since previous record is malformed
            {
                "data": {
                    "col1": "val_aa4_11",
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "a/a4.csv",
                },
                "stream": "stream1",
            },
            {
                "data": {
                    "col1": "val_bb1_11",
                    "col2": 1,
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "b/b1.csv",
                },
                "stream": "stream2",
            },
            {
                "data": {
                    "col1": "val_bb1_12",
                    "col2": 2,
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "b/b1.csv",
                },
                "stream": "stream2",
            },
            # {"data": {"col1": "val_bb2_11", "col2": "val_bb2_21", "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z", "_ab_source_file_url": "b/b2.csv"}, "stream": "stream2"},  # This record is skipped because it does not conform
            {
                "data": {
                    "col1": "val_bb2_12",
                    "col2": 2,
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "b/b2.csv",
                },
                "stream": "stream2",
            },
            {
                "data": {
                    "col1": "val_bb3_11",
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "b/b3.csv",
                },
                "stream": "stream2",
            },
            {
                "data": {
                    "col1": "val_bb3_12",
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "b/b3.csv",
                },
                "stream": "stream2",
            },
        ]
    )
    .set_expected_logs(
        {
            "read": [
                {
                    "level": "WARN",
                    "message": "Records in file did not pass validation policy. stream=stream1 file=a/a2.csv n_skipped=1 validation_policy=skip_record",
                },
                {
                    "level": "ERROR",
                    "message": "Error parsing record. This could be due to a mismatch between the config's file type and the actual file type, or because the file or record is not parseable. stream=stream1 file=a/a3.csv line_no=2 n_skipped=0",
                },
                {
                    "level": "WARN",
                    "message": "Records in file did not pass validation policy. stream=stream2 file=b/b2.csv n_skipped=1 validation_policy=skip_record",
                },
                {
                    "level": "WARN",
                    "message": "Could not cast the value to the expected type.: col2: value=this is text that will trigger validation policy,expected_type=integer",
                },
                {
                    "level": "WARN",
                    "message": "Could not cast the value to the expected type.: col2: value=this is text that will trigger validation policy,expected_type=integer",
                },
            ]
        }
    )
    .set_expected_read_error(
        AirbyteTracedException,
        "Please check the logged errors for more information.",
    )
).build()


emit_record_scenario_single_stream = (
    _base_single_stream_scenario.copy()
    .set_name("emit_record_scenario_single_stream")
    .set_config(
        {
            "streams": [
                {
                    "name": "stream1",
                    "format": {"filetype": "csv"},
                    "globs": ["*.csv"],
                    "validation_policy": "Emit Record",
                }
            ]
        }
    )
    .set_expected_records(
        [
            {
                "data": {
                    "col1": "val_a_11",
                    "col2": 1,
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "a.csv",
                },
                "stream": "stream1",
            },
            {
                "data": {
                    "col1": "val_a_12",
                    "col2": 2,
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "a.csv",
                },
                "stream": "stream1",
            },
            {
                "data": {
                    "col1": "val_b_11",
                    "col2": "this is text that will trigger validation policy",
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "b.csv",
                },
                "stream": "stream1",
            },  # This record is skipped because it does not conform
            {
                "data": {
                    "col1": "val_b_12",
                    "col2": 2,
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "b.csv",
                },
                "stream": "stream1",
            },
            {
                "data": {
                    "col1": "val_c_11",
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "c.csv",
                },
                "stream": "stream1",
            },
            # {"data": {"col1": "val_c_12", None: "val_c_22", "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z", "_ab_source_file_url": "c.csv"}, "stream": "stream1"},  # This record is malformed so should not be emitted
            # {"data": {"col1": "val_c_13", "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z", "_ab_source_file_url": "c.csv"}, "stream": "stream1"},  # No more records from this stream are emitted after we hit a parse error
            {
                "data": {
                    "col1": "val_d_11",
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "d.csv",
                },
                "stream": "stream1",
            },
        ]
    )
    .set_expected_read_error(
        AirbyteTracedException,
        "Please check the logged errors for more information.",
    )
).build()


emit_record_scenario_multi_stream = (
    _base_multi_stream_scenario.copy()
    .set_name("emit_record_scenario_multi_stream")
    .set_config(
        {
            "streams": [
                {
                    "name": "stream1",
                    "format": {"filetype": "csv"},
                    "globs": ["a/*.csv"],
                    "validation_policy": "Emit Record",
                },
                {
                    "name": "stream2",
                    "format": {"filetype": "csv"},
                    "globs": ["b/*.csv"],
                    "validation_policy": "Emit Record",
                },
            ]
        }
    )
    .set_expected_records(
        [
            {
                "data": {
                    "col1": "val_aa1_11",
                    "col2": 1,
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "a/a1.csv",
                },
                "stream": "stream1",
            },
            {
                "data": {
                    "col1": "val_aa1_12",
                    "col2": 2,
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "a/a1.csv",
                },
                "stream": "stream1",
            },
            {
                "data": {
                    "col1": "val_aa2_11",
                    "col2": "this is text that will trigger validation policy",
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "a/a2.csv",
                },
                "stream": "stream1",
            },
            {
                "data": {
                    "col1": "val_aa2_12",
                    "col2": 2,
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "a/a2.csv",
                },
                "stream": "stream1",
            },
            {
                "data": {
                    "col1": "val_aa3_11",
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "a/a3.csv",
                },
                "stream": "stream1",
            },
            # {"data": {"col1": "val_aa3_12", None: "val_aa3_22", "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z", "_ab_source_file_url": "a/a3.csv"}, "stream": "stream1"},  # This record is malformed so should not be emitted
            # {"data": {"col1": "val_aa3_13", "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z", "_ab_source_file_url": "a/a3.csv"}, "stream": "stream1"},  # Skipped since previous record is malformed
            {
                "data": {
                    "col1": "val_aa4_11",
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "a/a4.csv",
                },
                "stream": "stream1",
            },
            {
                "data": {
                    "col1": "val_bb1_11",
                    "col2": 1,
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "b/b1.csv",
                },
                "stream": "stream2",
            },
            {
                "data": {
                    "col1": "val_bb1_12",
                    "col2": 2,
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "b/b1.csv",
                },
                "stream": "stream2",
            },
            {
                "data": {
                    "col1": "val_bb2_11",
                    "col2": "this is text that will trigger validation policy",
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "b/b2.csv",
                },
                "stream": "stream2",
            },
            {
                "data": {
                    "col1": "val_bb2_12",
                    "col2": 2,
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "b/b2.csv",
                },
                "stream": "stream2",
            },
            {
                "data": {
                    "col1": "val_bb3_11",
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "b/b3.csv",
                },
                "stream": "stream2",
            },
            {
                "data": {
                    "col1": "val_bb3_12",
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "b/b3.csv",
                },
                "stream": "stream2",
            },
        ]
    )
    .set_expected_read_error(
        AirbyteTracedException,
        "Please check the logged errors for more information.",
    )
).build()


wait_for_rediscovery_scenario_single_stream = (
    _base_single_stream_scenario.copy()
    .set_name("wait_for_rediscovery_scenario_single_stream")
    .set_config(
        {
            "streams": [
                {
                    "name": "stream1",
                    "format": {"filetype": "csv"},
                    "globs": ["*.csv"],
                    "validation_policy": "Wait for Discover",
                }
            ]
        }
    )
    .set_expected_records(None)  # When syncing streams concurrently we don't know how many records will be emitted before the sync stops
    .set_expected_logs(
        {
            "read": [
                {
                    "level": "WARN",
                    "message": "Stopping sync in accordance with the configured validation policy. Records in file did not conform to the schema. stream=stream1 file=b.csv validation_policy=Wait for Discover n_skipped=0",
                },
                {
                    "level": "WARN",
                    "message": "Could not cast the value to the expected type.: col2: value=this is text that will trigger validation policy,expected_type=integer",
                },
            ]
        }
    )
).build()


wait_for_rediscovery_scenario_multi_stream = (
    _base_multi_stream_scenario.copy()
    .set_name("wait_for_rediscovery_scenario_multi_stream")
    .set_config(
        {
            "streams": [
                {
                    "name": "stream1",
                    "format": {"filetype": "csv"},
                    "globs": ["a/*.csv"],
                    "validation_policy": "Wait for Discover",
                },
                {
                    "name": "stream2",
                    "format": {"filetype": "csv"},
                    "globs": ["b/*.csv"],
                    "validation_policy": "Wait for Discover",
                },
            ]
        }
    )
    .set_expected_records(None)  # When syncing streams concurrently we don't know how many records will be emitted before the sync stops
    .set_expected_logs(
        {
            "read": [
                {
                    "level": "WARN",
                    "message": "Stopping sync in accordance with the configured validation policy. Records in file did not conform to the schema. stream=stream1 file=a/a2.csv validation_policy=Wait for Discover n_skipped=0",
                },
                {
                    "level": "WARN",
                    "message": "Stopping sync in accordance with the configured validation policy. Records in file did not conform to the schema. stream=stream2 file=b/b2.csv validation_policy=Wait for Discover n_skipped=0",
                },
                {
                    "level": "WARN",
                    "message": "Could not cast the value to the expected type.: col2: value=this is text that will trigger validation policy,expected_type=integer",
                },
                {
                    "level": "WARN",
                    "message": "Could not cast the value to the expected type.: col2: value=this is text that will trigger validation policy,expected_type=integer",
                },
            ]
        }
    )
).build()



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

import logging
import typing
from abc import ABC, abstractmethod
from typing import Optional, Tuple

from airbyte_cdk.sources.streams import Stream

if typing.TYPE_CHECKING:
    from airbyte_cdk.sources import Source


class AvailabilityStrategy(ABC):
    """
    Abstract base class for checking stream availability.
    """

    @abstractmethod
    def check_availability(self, stream: Stream, logger: logging.Logger, source: Optional["Source"]) -> Tuple[bool, Optional[str]]:
        """
        Checks stream availability.

        :param stream: stream
        :param logger: source logger
        :param source: (optional) source
        :return: A tuple of (boolean, str). If boolean is true, then the stream
          is available, and no str is required. Otherwise, the stream is unavailable
          for some reason and the str should describe what went wrong and how to
          resolve the unavailability, if possible.
        """



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

import json
from typing import Any, Iterator, List, Mapping
from unittest.mock import MagicMock, Mock, patch

import pytest
from airbyte_cdk.connector_builder.message_grouper import MessageGrouper
from airbyte_cdk.connector_builder.models import HttpRequest, HttpResponse, LogMessage, StreamRead, StreamReadPages
from airbyte_cdk.models import (
    AirbyteControlConnectorConfigMessage,
    AirbyteControlMessage,
    AirbyteLogMessage,
    AirbyteMessage,
    AirbyteRecordMessage,
    Level,
    OrchestratorType,
)
from airbyte_cdk.models import Type as MessageType
from unit_tests.connector_builder.utils import create_configured_catalog

_NO_PK = [[]]
_NO_CURSOR_FIELD = []

MAX_PAGES_PER_SLICE = 4
MAX_SLICES = 3

MANIFEST = {
    "version": "0.30.0",
    "type": "DeclarativeSource",
    "definitions": {
        "selector": {"extractor": {"field_path": ["items"], "type": "DpathExtractor"}, "type": "RecordSelector"},
        "requester": {"url_base": "https://demonslayers.com/api/v1/", "http_method": "GET", "type": "DeclarativeSource"},
        "retriever": {
            "type": "DeclarativeSource",
            "record_selector": {"extractor": {"field_path": ["items"], "type": "DpathExtractor"}, "type": "RecordSelector"},
            "paginator": {"type": "NoPagination"},
            "requester": {"url_base": "https://demonslayers.com/api/v1/", "http_method": "GET", "type": "HttpRequester"},
        },
        "hashiras_stream": {
            "retriever": {
                "type": "DeclarativeSource",
                "record_selector": {"extractor": {"field_path": ["items"], "type": "DpathExtractor"}, "type": "RecordSelector"},
                "paginator": {"type": "NoPagination"},
                "requester": {"url_base": "https://demonslayers.com/api/v1/", "http_method": "GET", "type": "HttpRequester"},
            },
            "$parameters": {"name": "hashiras", "path": "/hashiras"},
        },
        "breathing_techniques_stream": {
            "retriever": {
                "type": "DeclarativeSource",
                "record_selector": {"extractor": {"field_path": ["items"], "type": "DpathExtractor"}, "type": "RecordSelector"},
                "paginator": {"type": "NoPagination"},
                "requester": {"url_base": "https://demonslayers.com/api/v1/", "http_method": "GET", "type": "HttpRequester"},
            },
            "$parameters": {"name": "breathing-techniques", "path": "/breathing_techniques"},
        },
    },
    "streams": [
        {
            "type": "DeclarativeStream",
            "retriever": {
                "type": "SimpleRetriever",
                "record_selector": {"extractor": {"field_path": ["items"], "type": "DpathExtractor"}, "type": "RecordSelector"},
                "paginator": {"type": "NoPagination"},
                "requester": {"url_base": "https://demonslayers.com/api/v1/", "http_method": "GET", "type": "HttpRequester"},
            },
            "$parameters": {"name": "hashiras", "path": "/hashiras"},
        },
        {
            "type": "DeclarativeStream",
            "retriever": {
                "type": "SimpleRetriever",
                "record_selector": {"extractor": {"field_path": ["items"], "type": "DpathExtractor"}, "type": "RecordSelector"},
                "paginator": {"type": "NoPagination"},
                "requester": {"url_base": "https://demonslayers.com/api/v1/", "http_method": "GET", "type": "HttpRequester"},
            },
            "$parameters": {"name": "breathing-techniques", "path": "/breathing_techniques"},
        },
    ],
    "check": {"stream_names": ["hashiras"], "type": "CheckStream"},
}

CONFIG = {"rank": "upper-six"}

A_SOURCE = MagicMock()


@patch("airbyte_cdk.connector_builder.message_grouper.AirbyteEntrypoint.read")
def test_get_grouped_messages(mock_entrypoint_read: Mock) -> None:
    url = "https://demonslayers.com/api/v1/hashiras?era=taisho"
    request = {
        "headers": {"Content-Type": "application/json"},
        "method": "GET",
        "body": {"content": '{"custom": "field"}'},
    }
    response = {"status_code": 200, "headers": {"field": "value"}, "body": {"content": '{"name": "field"}'}}
    expected_schema = {
        "$schema": "http://json-schema.org/schema#",
        "properties": {"name": {"type": ["string", "null"]}, "date": {"type": ["string", "null"]}},
        "type": "object",
    }
    expected_datetime_fields = {"date": "%Y-%m-%d"}
    expected_pages = [
        StreamReadPages(
            request=HttpRequest(
                url="https://demonslayers.com/api/v1/hashiras",
                parameters={"era": ["taisho"]},
                headers={"Content-Type": "application/json"},
                body='{"custom": "field"}',
                http_method="GET",
            ),
            response=HttpResponse(status=200, headers={"field": "value"}, body='{"name": "field"}'),
            records=[{"name": "Shinobu Kocho", "date": "2023-03-03"}, {"name": "Muichiro Tokito", "date": "2023-03-04"}],
        ),
        StreamReadPages(
            request=HttpRequest(
                url="https://demonslayers.com/api/v1/hashiras",
                parameters={"era": ["taisho"]},
                headers={"Content-Type": "application/json"},
                body='{"custom": "field"}',
                http_method="GET",
            ),
            response=HttpResponse(status=200, headers={"field": "value"}, body='{"name": "field"}'),
            records=[{"name": "Mitsuri Kanroji", "date": "2023-03-05"}],
        ),
    ]

    mock_source = make_mock_source(
        mock_entrypoint_read,
        iter(
            [
                request_response_log_message(request, response, url),
                record_message("hashiras", {"name": "Shinobu Kocho", "date": "2023-03-03"}),
                record_message("hashiras", {"name": "Muichiro Tokito", "date": "2023-03-04"}),
                request_response_log_message(request, response, url),
                record_message("hashiras", {"name": "Mitsuri Kanroji", "date": "2023-03-05"}),
            ]
        ),
    )

    connector_builder_handler = MessageGrouper(MAX_PAGES_PER_SLICE, MAX_SLICES)
    actual_response: StreamRead = connector_builder_handler.get_message_groups(
        source=mock_source, config=CONFIG, configured_catalog=create_configured_catalog("hashiras")
    )

    assert actual_response.inferred_schema == expected_schema
    assert actual_response.inferred_datetime_formats == expected_datetime_fields

    single_slice = actual_response.slices[0]
    for i, actual_page in enumerate(single_slice.pages):
        assert actual_page == expected_pages[i]


@patch("airbyte_cdk.connector_builder.message_grouper.AirbyteEntrypoint.read")
def test_get_grouped_messages_with_logs(mock_entrypoint_read: Mock) -> None:
    url = "https://demonslayers.com/api/v1/hashiras?era=taisho"
    request = {
        "headers": {"Content-Type": "application/json"},
        "method": "GET",
        "body": {"content": '{"custom": "field"}'},
    }
    response = {"status_code": 200, "headers": {"field": "value"}, "body": {"content": '{"name": "field"}'}}
    expected_pages = [
        StreamReadPages(
            request=HttpRequest(
                url="https://demonslayers.com/api/v1/hashiras",
                parameters={"era": ["taisho"]},
                headers={"Content-Type": "application/json"},
                body='{"custom": "field"}',
                http_method="GET",
            ),
            response=HttpResponse(status=200, headers={"field": "value"}, body='{"name": "field"}'),
            records=[{"name": "Shinobu Kocho"}, {"name": "Muichiro Tokito"}],
        ),
        StreamReadPages(
            request=HttpRequest(
                url="https://demonslayers.com/api/v1/hashiras",
                parameters={"era": ["taisho"]},
                headers={"Content-Type": "application/json"},
                body='{"custom": "field"}',
                http_method="GET",
            ),
            response=HttpResponse(status=200, headers={"field": "value"}, body='{"name": "field"}'),
            records=[{"name": "Mitsuri Kanroji"}],
        ),
    ]
    expected_logs = [
        LogMessage(**{"message": "log message before the request", "level": "INFO"}),
        LogMessage(**{"message": "log message during the page", "level": "INFO"}),
        LogMessage(**{"message": "log message after the response", "level": "INFO"}),
    ]

    mock_source = make_mock_source(
        mock_entrypoint_read,
        iter(
            [
                AirbyteMessage(type=MessageType.LOG, log=AirbyteLogMessage(level=Level.INFO, message="log message before the request")),
                request_response_log_message(request, response, url),
                record_message("hashiras", {"name": "Shinobu Kocho"}),
                AirbyteMessage(type=MessageType.LOG, log=AirbyteLogMessage(level=Level.INFO, message="log message during the page")),
                record_message("hashiras", {"name": "Muichiro Tokito"}),
                AirbyteMessage(type=MessageType.LOG, log=AirbyteLogMessage(level=Level.INFO, message="log message after the response")),
            ]
        ),
    )

    connector_builder_handler = MessageGrouper(MAX_PAGES_PER_SLICE, MAX_SLICES)

    actual_response: StreamRead = connector_builder_handler.get_message_groups(
        source=mock_source, config=CONFIG, configured_catalog=create_configured_catalog("hashiras")
    )
    single_slice = actual_response.slices[0]
    for i, actual_page in enumerate(single_slice.pages):
        assert actual_page == expected_pages[i]

    for i, actual_log in enumerate(actual_response.logs):
        assert actual_log == expected_logs[i]


@pytest.mark.parametrize(
    "request_record_limit, max_record_limit, should_fail",
    [
        pytest.param(1, 3, False, id="test_create_request_with_record_limit"),
        pytest.param(3, 1, True, id="test_create_request_record_limit_exceeds_max"),
    ],
)
@patch("airbyte_cdk.connector_builder.message_grouper.AirbyteEntrypoint.read")
def test_get_grouped_messages_record_limit(mock_entrypoint_read: Mock, request_record_limit: int, max_record_limit: int, should_fail: bool) -> None:
    url = "https://demonslayers.com/api/v1/hashiras?era=taisho"
    request = {
        "headers": {"Content-Type": "application/json"},
        "method": "GET",
        "body": {"content": '{"custom": "field"}'},
    }
    response = {"status_code": 200, "headers": {"field": "value"}, "body": {"content": '{"name": "field"}'}}
    mock_source = make_mock_source(
        mock_entrypoint_read,
        iter(
            [
                request_response_log_message(request, response, url),
                record_message("hashiras", {"name": "Shinobu Kocho"}),
                record_message("hashiras", {"name": "Muichiro Tokito"}),
                request_response_log_message(request, response, url),
                record_message("hashiras", {"name": "Mitsuri Kanroji"}),
            ]
        ),
    )
    n_records = 2
    record_limit = min(request_record_limit, max_record_limit)

    api = MessageGrouper(MAX_PAGES_PER_SLICE, MAX_SLICES, max_record_limit=max_record_limit)
    # this is the call we expect to raise an exception
    if should_fail:
        with pytest.raises(ValueError):
            api.get_message_groups(
                mock_source, config=CONFIG, configured_catalog=create_configured_catalog("hashiras"), record_limit=request_record_limit
            )
    else:
        actual_response: StreamRead = api.get_message_groups(
            mock_source, config=CONFIG, configured_catalog=create_configured_catalog("hashiras"), record_limit=request_record_limit
        )
        single_slice = actual_response.slices[0]
        total_records = 0
        for i, actual_page in enumerate(single_slice.pages):
            total_records += len(actual_page.records)
        assert total_records == min([record_limit, n_records])

        assert (total_records >= max_record_limit) == actual_response.test_read_limit_reached


@pytest.mark.parametrize(
    "max_record_limit",
    [
        pytest.param(2, id="test_create_request_no_record_limit"),
        pytest.param(1, id="test_create_request_no_record_limit_n_records_exceed_max"),
    ],
)
@patch("airbyte_cdk.connector_builder.message_grouper.AirbyteEntrypoint.read")
def test_get_grouped_messages_default_record_limit(mock_entrypoint_read: Mock, max_record_limit: int) -> None:
    url = "https://demonslayers.com/api/v1/hashiras?era=taisho"
    request = {
        "headers": {"Content-Type": "application/json"},
        "method": "GET",
        "body": {"content": '{"custom": "field"}'},
    }
    response = {"status_code": 200, "headers": {"field": "value"}, "body": {"content": '{"name": "field"}'}}
    mock_source = make_mock_source(
        mock_entrypoint_read,
        iter(
            [
                request_response_log_message(request, response, url),
                record_message("hashiras", {"name": "Shinobu Kocho"}),
                record_message("hashiras", {"name": "Muichiro Tokito"}),
                request_response_log_message(request, response, url),
                record_message("hashiras", {"name": "Mitsuri Kanroji"}),
            ]
        ),
    )
    n_records = 2

    api = MessageGrouper(MAX_PAGES_PER_SLICE, MAX_SLICES, max_record_limit=max_record_limit)
    actual_response: StreamRead = api.get_message_groups(
        source=mock_source, config=CONFIG, configured_catalog=create_configured_catalog("hashiras")
    )
    single_slice = actual_response.slices[0]
    total_records = 0
    for i, actual_page in enumerate(single_slice.pages):
        total_records += len(actual_page.records)
    assert total_records == min([max_record_limit, n_records])


@patch("airbyte_cdk.connector_builder.message_grouper.AirbyteEntrypoint.read")
def test_get_grouped_messages_limit_0(mock_entrypoint_read: Mock) -> None:
    url = "https://demonslayers.com/api/v1/hashiras?era=taisho"
    request = {
        "headers": {"Content-Type": "application/json"},
        "method": "GET",
        "body": {"content": '{"custom": "field"}'},
    }
    response = {"status_code": 200, "headers": {"field": "value"}, "body": {"content": '{"name": "field"}'}}
    mock_source = make_mock_source(
        mock_entrypoint_read,
        iter(
            [
                request_response_log_message(request, response, url),
                record_message("hashiras", {"name": "Shinobu Kocho"}),
                record_message("hashiras", {"name": "Muichiro Tokito"}),
                request_response_log_message(request, response, url),
                record_message("hashiras", {"name": "Mitsuri Kanroji"}),
            ]
        ),
    )
    api = MessageGrouper(MAX_PAGES_PER_SLICE, MAX_SLICES)

    with pytest.raises(ValueError):
        api.get_message_groups(source=mock_source, config=CONFIG, configured_catalog=create_configured_catalog("hashiras"), record_limit=0)


@patch("airbyte_cdk.connector_builder.message_grouper.AirbyteEntrypoint.read")
def test_get_grouped_messages_no_records(mock_entrypoint_read: Mock) -> None:
    url = "https://demonslayers.com/api/v1/hashiras?era=taisho"
    request = {
        "headers": {"Content-Type": "application/json"},
        "method": "GET",
        "body": {"content": '{"custom": "field"}'},
    }
    response = {"status_code": 200, "headers": {"field": "value"}, "body": {"content": '{"name": "field"}'}}
    expected_pages = [
        StreamReadPages(
            request=HttpRequest(
                url="https://demonslayers.com/api/v1/hashiras",
                parameters={"era": ["taisho"]},
                headers={"Content-Type": "application/json"},
                body='{"custom": "field"}',
                http_method="GET",
            ),
            response=HttpResponse(status=200, headers={"field": "value"}, body='{"name": "field"}'),
            records=[],
        ),
        StreamReadPages(
            request=HttpRequest(
                url="https://demonslayers.com/api/v1/hashiras",
                parameters={"era": ["taisho"]},
                headers={"Content-Type": "application/json"},
                body='{"custom": "field"}',
                http_method="GET",
            ),
            response=HttpResponse(status=200, headers={"field": "value"}, body='{"name": "field"}'),
            records=[],
        ),
    ]

    mock_source = make_mock_source(
        mock_entrypoint_read,
        iter(
            [
                request_response_log_message(request, response, url),
                request_response_log_message(request, response, url),
            ]
        ),
    )

    message_grouper = MessageGrouper(MAX_PAGES_PER_SLICE, MAX_SLICES)

    actual_response: StreamRead = message_grouper.get_message_groups(
        source=mock_source, config=CONFIG, configured_catalog=create_configured_catalog("hashiras")
    )

    single_slice = actual_response.slices[0]
    for i, actual_page in enumerate(single_slice.pages):
        assert actual_page == expected_pages[i]


@pytest.mark.parametrize(
    "log_message, expected_response",
    [
        pytest.param(
            {
                "http": {
                    "response": {
                        "status_code": 200,
                        "headers": {"field": "name"},
                        "body": {"content": '{"id": "fire", "owner": "kyojuro_rengoku"}'},
                    }
                }
            },
            HttpResponse(status=200, headers={"field": "name"}, body='{"id": "fire", "owner": "kyojuro_rengoku"}'),
            id="test_create_response_with_all_fields",
        ),
        pytest.param(
            {"http": {"response": {"status_code": 200, "headers": {"field": "name"}}}},
            HttpResponse(status=200, headers={"field": "name"}, body=""),
            id="test_create_response_with_no_body",
        ),
        pytest.param(
            {"http": {"response": {"status_code": 200, "body": {"content": '{"id": "fire", "owner": "kyojuro_rengoku"}'}}}},
            HttpResponse(status=200, body='{"id": "fire", "owner": "kyojuro_rengoku"}'),
            id="test_create_response_with_no_headers",
        ),
        pytest.param(
            {
                "http": {
                    "response": {
                        "status_code": 200,
                        "headers": {"field": "name"},
                        "body": {"content": '[{"id": "fire", "owner": "kyojuro_rengoku"}, {"id": "mist", "owner": "muichiro_tokito"}]'},
                    }
                }
            },
            HttpResponse(
                status=200,
                headers={"field": "name"},
                body='[{"id": "fire", "owner": "kyojuro_rengoku"}, {"id": "mist", "owner": "muichiro_tokito"}]',
            ),
            id="test_create_response_with_array",
        ),
        pytest.param(
            {"http": {"response": {"status_code": 200, "body": {"content": "tomioka"}}}},
            HttpResponse(status=200, body="tomioka"),
            id="test_create_response_with_string",
        ),
    ],
)
def test_create_response_from_log_message(log_message: str, expected_response: HttpResponse) -> None:
    if isinstance(log_message, str):
        response_message = json.loads(log_message)
    else:
        response_message = log_message

    connector_builder_handler = MessageGrouper(MAX_PAGES_PER_SLICE, MAX_SLICES)
    actual_response = connector_builder_handler._create_response_from_log_message(response_message)

    assert actual_response == expected_response


@patch("airbyte_cdk.connector_builder.message_grouper.AirbyteEntrypoint.read")
def test_get_grouped_messages_with_many_slices(mock_entrypoint_read: Mock) -> None:
    url = "http://a-url.com"
    request: Mapping[str, Any] = {}
    response = {"status_code": 200}

    mock_source = make_mock_source(
        mock_entrypoint_read,
        iter(
            [
                slice_message('{"descriptor": "first_slice"}'),
                request_response_log_message(request, response, url),
                record_message("hashiras", {"name": "Muichiro Tokito"}),
                slice_message('{"descriptor": "second_slice"}'),
                request_response_log_message(request, response, url),
                record_message("hashiras", {"name": "Shinobu Kocho"}),
                record_message("hashiras", {"name": "Mitsuri Kanroji"}),
                request_response_log_message(request, response, url),
                record_message("hashiras", {"name": "Obanai Iguro"}),
                request_response_log_message(request, response, url),
            ]
        ),
    )

    connector_builder_handler = MessageGrouper(MAX_PAGES_PER_SLICE, MAX_SLICES)

    stream_read: StreamRead = connector_builder_handler.get_message_groups(
        source=mock_source, config=CONFIG, configured_catalog=create_configured_catalog("hashiras")
    )

    assert not stream_read.test_read_limit_reached
    assert len(stream_read.slices) == 2

    assert stream_read.slices[0].slice_descriptor == {"descriptor": "first_slice"}
    assert len(stream_read.slices[0].pages) == 1
    assert len(stream_read.slices[0].pages[0].records) == 1

    assert stream_read.slices[1].slice_descriptor == {"descriptor": "second_slice"}
    assert len(stream_read.slices[1].pages) == 3
    assert len(stream_read.slices[1].pages[0].records) == 2
    assert len(stream_read.slices[1].pages[1].records) == 1
    assert len(stream_read.slices[1].pages[2].records) == 0


@patch("airbyte_cdk.connector_builder.message_grouper.AirbyteEntrypoint.read")
def test_get_grouped_messages_given_maximum_number_of_slices_then_test_read_limit_reached(mock_entrypoint_read: Mock) -> None:
    maximum_number_of_slices = 5
    request: Mapping[str, Any] = {}
    response = {"status_code": 200}
    mock_source = make_mock_source(
        mock_entrypoint_read, iter([slice_message(), request_response_log_message(request, response, "a_url")] * maximum_number_of_slices)
    )

    api = MessageGrouper(MAX_PAGES_PER_SLICE, MAX_SLICES)

    stream_read: StreamRead = api.get_message_groups(
        source=mock_source, config=CONFIG, configured_catalog=create_configured_catalog("hashiras")
    )

    assert stream_read.test_read_limit_reached


@patch("airbyte_cdk.connector_builder.message_grouper.AirbyteEntrypoint.read")
def test_get_grouped_messages_given_maximum_number_of_pages_then_test_read_limit_reached(mock_entrypoint_read: Mock) -> None:
    maximum_number_of_pages_per_slice = 5
    request: Mapping[str, Any] = {}
    response = {"status_code": 200}
    mock_source = make_mock_source(
        mock_entrypoint_read,
        iter([slice_message()] + [request_response_log_message(request, response, "a_url")] * maximum_number_of_pages_per_slice),
    )

    api = MessageGrouper(MAX_PAGES_PER_SLICE, MAX_SLICES)

    stream_read: StreamRead = api.get_message_groups(
        source=mock_source, config=CONFIG, configured_catalog=create_configured_catalog("hashiras")
    )

    assert stream_read.test_read_limit_reached


def test_read_stream_returns_error_if_stream_does_not_exist() -> None:
    mock_source = MagicMock()
    mock_source.read.side_effect = ValueError("error")
    mock_source.streams.return_value = [make_mock_stream()]

    full_config: Mapping[str, Any] = {**CONFIG, **{"__injected_declarative_manifest": MANIFEST}}

    message_grouper = MessageGrouper(MAX_PAGES_PER_SLICE, MAX_SLICES)
    actual_response = message_grouper.get_message_groups(
        source=mock_source, config=full_config, configured_catalog=create_configured_catalog("not_in_manifest")
    )

    assert 1 == len(actual_response.logs)
    assert "Traceback" in actual_response.logs[0].message
    assert "ERROR" in actual_response.logs[0].level


@patch("airbyte_cdk.connector_builder.message_grouper.AirbyteEntrypoint.read")
def test_given_control_message_then_stream_read_has_config_update(mock_entrypoint_read: Mock) -> None:
    updated_config = {"x": 1}
    mock_source = make_mock_source(
        mock_entrypoint_read, iter(any_request_and_response_with_a_record() + [connector_configuration_control_message(1, updated_config)])
    )
    connector_builder_handler = MessageGrouper(MAX_PAGES_PER_SLICE, MAX_SLICES)
    stream_read: StreamRead = connector_builder_handler.get_message_groups(
        source=mock_source, config=CONFIG, configured_catalog=create_configured_catalog("hashiras")
    )

    assert stream_read.latest_config_update == updated_config


@patch("airbyte_cdk.connector_builder.message_grouper.AirbyteEntrypoint.read")
def test_given_multiple_control_messages_then_stream_read_has_latest_based_on_emitted_at(mock_entrypoint_read: Mock) -> None:
    earliest = 0
    earliest_config = {"earliest": 0}
    latest = 1
    latest_config = {"latest": 1}
    mock_source = make_mock_source(
        mock_entrypoint_read,
        iter(
            any_request_and_response_with_a_record()
            + [
                # here, we test that even if messages are emitted in a different order, we still rely on `emitted_at`
                connector_configuration_control_message(latest, latest_config),
                connector_configuration_control_message(earliest, earliest_config),
            ]
        ),
    )
    connector_builder_handler = MessageGrouper(MAX_PAGES_PER_SLICE, MAX_SLICES)
    stream_read: StreamRead = connector_builder_handler.get_message_groups(
        source=mock_source, config=CONFIG, configured_catalog=create_configured_catalog("hashiras")
    )

    assert stream_read.latest_config_update == latest_config


@patch("airbyte_cdk.connector_builder.message_grouper.AirbyteEntrypoint.read")
def test_given_multiple_control_messages_with_same_timestamp_then_stream_read_has_latest_based_on_message_order(
    mock_entrypoint_read: Mock,
) -> None:
    emitted_at = 0
    earliest_config = {"earliest": 0}
    latest_config = {"latest": 1}
    mock_source = make_mock_source(
        mock_entrypoint_read,
        iter(
            any_request_and_response_with_a_record()
            + [
                connector_configuration_control_message(emitted_at, earliest_config),
                connector_configuration_control_message(emitted_at, latest_config),
            ]
        ),
    )
    connector_builder_handler = MessageGrouper(MAX_PAGES_PER_SLICE, MAX_SLICES)
    stream_read: StreamRead = connector_builder_handler.get_message_groups(
        source=mock_source, config=CONFIG, configured_catalog=create_configured_catalog("hashiras")
    )

    assert stream_read.latest_config_update == latest_config


@patch("airbyte_cdk.connector_builder.message_grouper.AirbyteEntrypoint.read")
def test_given_auxiliary_requests_then_return_auxiliary_request(mock_entrypoint_read: Mock) -> None:
    mock_source = make_mock_source(mock_entrypoint_read, iter(any_request_and_response_with_a_record() + [auxiliary_request_log_message()]))
    connector_builder_handler = MessageGrouper(MAX_PAGES_PER_SLICE, MAX_SLICES)
    stream_read: StreamRead = connector_builder_handler.get_message_groups(
        source=mock_source, config=CONFIG, configured_catalog=create_configured_catalog("hashiras")
    )

    assert len(stream_read.auxiliary_requests) == 1


@patch("airbyte_cdk.connector_builder.message_grouper.AirbyteEntrypoint.read")
def test_given_no_slices_then_return_empty_slices(mock_entrypoint_read: Mock) -> None:
    mock_source = make_mock_source(mock_entrypoint_read, iter([auxiliary_request_log_message()]))
    connector_builder_handler = MessageGrouper(MAX_PAGES_PER_SLICE, MAX_SLICES)
    stream_read: StreamRead = connector_builder_handler.get_message_groups(
        source=mock_source, config=CONFIG, configured_catalog=create_configured_catalog("hashiras")
    )

    assert len(stream_read.slices) == 0


@patch("airbyte_cdk.connector_builder.message_grouper.AirbyteEntrypoint.read")
def test_given_pk_then_ensure_pk_is_pass_to_schema_inferrence(mock_entrypoint_read: Mock) -> None:
    mock_source = make_mock_source(mock_entrypoint_read, iter([
        request_response_log_message({"request": 1}, {"response": 2}, "http://any_url.com"),
        record_message("hashiras", {"id": "Shinobu Kocho", "date": "2023-03-03"}),
        record_message("hashiras", {"id": "Muichiro Tokito", "date": "2023-03-04"}),
    ]))
    mock_source.streams.return_value = [Mock()]
    mock_source.streams.return_value[0].primary_key = [["id"]]
    mock_source.streams.return_value[0].cursor_field = _NO_CURSOR_FIELD
    connector_builder_handler = MessageGrouper(MAX_PAGES_PER_SLICE, MAX_SLICES)

    stream_read: StreamRead = connector_builder_handler.get_message_groups(
        source=mock_source, config=CONFIG, configured_catalog=create_configured_catalog("hashiras")
    )

    assert stream_read.inferred_schema["required"] == ["id"]


@patch("airbyte_cdk.connector_builder.message_grouper.AirbyteEntrypoint.read")
def test_given_cursor_field_then_ensure_cursor_field_is_pass_to_schema_inferrence(mock_entrypoint_read: Mock) -> None:
    mock_source = make_mock_source(mock_entrypoint_read, iter([
        request_response_log_message({"request": 1}, {"response": 2}, "http://any_url.com"),
        record_message("hashiras", {"id": "Shinobu Kocho", "date": "2023-03-03"}),
        record_message("hashiras", {"id": "Muichiro Tokito", "date": "2023-03-04"}),
    ]))
    mock_source.streams.return_value = [Mock()]
    mock_source.streams.return_value[0].primary_key = _NO_PK
    mock_source.streams.return_value[0].cursor_field = ["date"]
    connector_builder_handler = MessageGrouper(MAX_PAGES_PER_SLICE, MAX_SLICES)

    stream_read: StreamRead = connector_builder_handler.get_message_groups(
        source=mock_source, config=CONFIG, configured_catalog=create_configured_catalog("hashiras")
    )

    assert stream_read.inferred_schema["required"] == ["date"]


def make_mock_source(mock_entrypoint_read: Mock, return_value: Iterator[AirbyteMessage]) -> MagicMock:
    mock_source = MagicMock()
    mock_entrypoint_read.return_value = return_value
    mock_source.streams.return_value = [make_mock_stream()]
    return mock_source


def make_mock_stream():
    mock_stream = MagicMock()
    mock_stream.primary_key = []
    mock_stream.cursor_field = []
    return mock_stream


def request_log_message(request: Mapping[str, Any]) -> AirbyteMessage:
    return AirbyteMessage(type=MessageType.LOG, log=AirbyteLogMessage(level=Level.INFO, message=f"request:{json.dumps(request)}"))


def response_log_message(response: Mapping[str, Any]) -> AirbyteMessage:
    return AirbyteMessage(type=MessageType.LOG, log=AirbyteLogMessage(level=Level.INFO, message=f"response:{json.dumps(response)}"))


def record_message(stream: str, data: Mapping[str, Any]) -> AirbyteMessage:
    return AirbyteMessage(type=MessageType.RECORD, record=AirbyteRecordMessage(stream=stream, data=data, emitted_at=1234))


def slice_message(slice_descriptor: str = '{"key": "value"}') -> AirbyteMessage:
    return AirbyteMessage(type=MessageType.LOG, log=AirbyteLogMessage(level=Level.INFO, message="slice:" + slice_descriptor))


def connector_configuration_control_message(emitted_at: float, config: Mapping[str, Any]) -> AirbyteMessage:
    return AirbyteMessage(
        type=MessageType.CONTROL,
        control=AirbyteControlMessage(
            type=OrchestratorType.CONNECTOR_CONFIG,
            emitted_at=emitted_at,
            connectorConfig=AirbyteControlConnectorConfigMessage(config=config),
        ),
    )


def auxiliary_request_log_message() -> AirbyteMessage:
    return AirbyteMessage(
        type=MessageType.LOG,
        log=AirbyteLogMessage(
            level=Level.INFO,
            message=json.dumps(
                {
                    "http": {
                        "is_auxiliary": True,
                        "title": "a title",
                        "description": "a description",
                        "request": {},
                        "response": {},
                    },
                    "url": {"full": "https://a-url.com"},
                }
            ),
        ),
    )


def request_response_log_message(request: Mapping[str, Any], response: Mapping[str, Any], url: str) -> AirbyteMessage:
    return AirbyteMessage(
        type=MessageType.LOG,
        log=AirbyteLogMessage(
            level=Level.INFO,
            message=json.dumps(
                {
                    "airbyte_cdk": {"stream": {"name": "a stream name"}},
                    "http": {"title": "a title", "description": "a description", "request": request, "response": response},
                    "url": {"full": url},
                }
            ),
        ),
    )


def any_request_and_response_with_a_record() -> List[AirbyteMessage]:
    return [
        request_response_log_message({"request": 1}, {"response": 2}, "http://any_url.com"),
        record_message("hashiras", {"name": "Shinobu Kocho"}),
    ]



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

import asyncio
import io
import json
from typing import Any, Dict
from unittest.mock import MagicMock, Mock

import pytest
from airbyte_cdk.sources.file_based.exceptions import RecordParseError
from airbyte_cdk.sources.file_based.file_based_stream_reader import AbstractFileBasedStreamReader
from airbyte_cdk.sources.file_based.file_types import JsonlParser

JSONL_CONTENT_WITHOUT_MULTILINE_JSON_OBJECTS = [
    b'{"a": 1, "b": "1"}',
    b'{"a": 2, "b": "2"}',
]
JSONL_CONTENT_WITH_MULTILINE_JSON_OBJECTS = [
    b"{",
    b'  "a": 1,',
    b'  "b": "1"',
    b"}",
    b"{",
    b'  "a": 2,',
    b'  "b": "2"',
    b"}",
]
INVALID_JSON_CONTENT = [
    b"{",
    b'  "a": 1,',
    b'  "b": "1"',
    b"{",
    b'  "a": 2,',
    b'  "b": "2"',
    b"}",
]


@pytest.fixture
def stream_reader() -> MagicMock:
    return MagicMock(spec=AbstractFileBasedStreamReader)


def _infer_schema(stream_reader: MagicMock) -> Dict[str, Any]:
    loop = asyncio.new_event_loop()
    task = loop.create_task(JsonlParser().infer_schema(Mock(), Mock(), stream_reader, Mock()))
    loop.run_until_complete(task)
    return task.result()  # type: ignore  # asyncio has no typing


def test_when_infer_then_return_proper_types(stream_reader: MagicMock) -> None:
    record = {"col1": 1, "col2": 2.2, "col3": "3", "col4": ["a", "list"], "col5": {"inner": "obj"}, "col6": None, "col7": True}
    stream_reader.open_file.return_value.__enter__.return_value = io.BytesIO(json.dumps(record).encode("utf-8"))

    schema = _infer_schema(stream_reader)

    assert schema == {
        "col1": {"type": "integer"},
        "col2": {"type": "number"},
        "col3": {"type": "string"},
        "col4": {"type": "array"},
        "col5": {"type": "object"},
        "col6": {"type": "null"},
        "col7": {"type": "boolean"},
    }


def test_given_str_io_when_infer_then_return_proper_types(stream_reader: MagicMock) -> None:
    stream_reader.open_file.return_value.__enter__.return_value = io.StringIO('{"col": 1}')

    schema = _infer_schema(stream_reader)

    assert schema == {"col": {"type": "integer"}}


def test_given_empty_record_when_infer_then_return_empty_schema(stream_reader: MagicMock) -> None:
    stream_reader.open_file.return_value.__enter__.return_value = io.BytesIO("{}".encode("utf-8"))
    schema = _infer_schema(stream_reader)
    assert schema == {}


def test_given_no_records_when_infer_then_return_empty_schema(stream_reader: MagicMock) -> None:
    stream_reader.open_file.return_value.__enter__.return_value = io.BytesIO("".encode("utf-8"))
    schema = _infer_schema(stream_reader)
    assert schema == {}


def test_given_limit_hit_when_infer_then_stop_considering_records(stream_reader: MagicMock) -> None:
    jsonl_file_content = '{"key": 2.' + "2" * JsonlParser.MAX_BYTES_PER_FILE_FOR_SCHEMA_INFERENCE + '}\n{"key": "a string"}'
    stream_reader.open_file.return_value.__enter__.return_value = io.BytesIO(jsonl_file_content.encode("utf-8"))

    schema = _infer_schema(stream_reader)

    assert schema == {"key": {"type": "number"}}


def test_given_multiline_json_objects_and_read_limit_hit_when_infer_then_return_parse_until_at_least_one_record(
    stream_reader: MagicMock,
) -> None:
    jsonl_file_content = '{\n"key": 2.' + "2" * JsonlParser.MAX_BYTES_PER_FILE_FOR_SCHEMA_INFERENCE + "\n}"
    stream_reader.open_file.return_value.__enter__.return_value = io.BytesIO(jsonl_file_content.encode("utf-8"))

    schema = _infer_schema(stream_reader)

    assert schema == {"key": {"type": "number"}}


def test_given_multiline_json_objects_and_hits_read_limit_when_infer_then_return_proper_types(stream_reader: MagicMock) -> None:
    stream_reader.open_file.return_value.__enter__.return_value = JSONL_CONTENT_WITH_MULTILINE_JSON_OBJECTS
    schema = _infer_schema(stream_reader)
    assert schema == {"a": {"type": "integer"}, "b": {"type": "string"}}


def test_given_multiple_records_then_merge_types(stream_reader: MagicMock) -> None:
    stream_reader.open_file.return_value.__enter__.return_value = io.BytesIO('{"col1": 1}\n{"col1": 2.3}'.encode("utf-8"))
    schema = _infer_schema(stream_reader)
    assert schema == {"col1": {"type": "number"}}


def test_given_one_json_per_line_when_parse_records_then_return_records(stream_reader: MagicMock) -> None:
    stream_reader.open_file.return_value.__enter__.return_value = JSONL_CONTENT_WITHOUT_MULTILINE_JSON_OBJECTS
    records = list(JsonlParser().parse_records(Mock(), Mock(), stream_reader, Mock(), None))
    assert records == [{"a": 1, "b": "1"}, {"a": 2, "b": "2"}]


def test_given_one_json_per_line_when_parse_records_then_do_not_send_warning(stream_reader: MagicMock) -> None:
    stream_reader.open_file.return_value.__enter__.return_value = JSONL_CONTENT_WITHOUT_MULTILINE_JSON_OBJECTS
    logger = Mock()

    list(JsonlParser().parse_records(Mock(), Mock(), stream_reader, logger, None))

    assert logger.warning.call_count == 0


def test_given_multiline_json_object_when_parse_records_then_return_records(stream_reader: MagicMock) -> None:
    stream_reader.open_file.return_value.__enter__.return_value = JSONL_CONTENT_WITH_MULTILINE_JSON_OBJECTS
    records = list(JsonlParser().parse_records(Mock(), Mock(), stream_reader, Mock(), None))
    assert records == [{"a": 1, "b": "1"}, {"a": 2, "b": "2"}]


def test_given_multiline_json_object_when_parse_records_then_log_once_one_record_yielded(stream_reader: MagicMock) -> None:
    stream_reader.open_file.return_value.__enter__.return_value = JSONL_CONTENT_WITH_MULTILINE_JSON_OBJECTS
    logger = Mock()

    next(iter(JsonlParser().parse_records(Mock(), Mock(), stream_reader, logger, None)))

    assert logger.warning.call_count == 1


def test_given_unparsable_json_when_parse_records_then_raise_error(stream_reader: MagicMock) -> None:
    stream_reader.open_file.return_value.__enter__.return_value = INVALID_JSON_CONTENT
    logger = Mock()

    with pytest.raises(RecordParseError):
        list(JsonlParser().parse_records(Mock(), Mock(), stream_reader, logger, None))
    assert logger.warning.call_count == 0



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#
import concurrent
import logging
from typing import Any, Callable, Dict, Iterable, Mapping, Optional, Tuple
from unittest.mock import Mock

from airbyte_cdk.models import SyncMode
from airbyte_cdk.sources.concurrent_source.concurrent_source import ConcurrentSource
from airbyte_cdk.sources.concurrent_source.thread_pool_manager import ThreadPoolManager
from airbyte_cdk.sources.message import InMemoryMessageRepository, MessageRepository
from airbyte_cdk.sources.streams.concurrent.abstract_stream import AbstractStream
from airbyte_cdk.sources.streams.concurrent.availability_strategy import StreamAvailability, StreamAvailable, StreamUnavailable
from airbyte_cdk.sources.streams.concurrent.cursor import Cursor, FinalStateCursor
from airbyte_cdk.sources.streams.concurrent.partitions.partition import Partition
from airbyte_cdk.sources.streams.concurrent.partitions.record import Record
from airbyte_protocol.models import AirbyteStream

logger = logging.getLogger("airbyte")


class _MockSource(ConcurrentSource):
    def __init__(
        self,
        check_lambda: Callable[[], Tuple[bool, Optional[Any]]] = None,
        per_stream: bool = True,
        message_repository: MessageRepository = InMemoryMessageRepository(),
        threadpool: ThreadPoolManager = ThreadPoolManager(
            concurrent.futures.ThreadPoolExecutor(max_workers=1, thread_name_prefix="workerpool"), logger
        ),
        exception_on_missing_stream: bool = True,
    ):
        super().__init__(threadpool, Mock(), Mock(), message_repository)
        self.check_lambda = check_lambda
        self.per_stream = per_stream
        self.exception_on_missing_stream = exception_on_missing_stream
        self._message_repository = message_repository


MESSAGE_FROM_REPOSITORY = Mock()


class _MockStream(AbstractStream):
    def __init__(self, name: str, message_repository: MessageRepository, available: bool = True, json_schema: Dict[str, Any] = {}):
        self._name = name
        self._available = available
        self._json_schema = json_schema
        self._message_repository = message_repository

    def generate_partitions(self) -> Iterable[Partition]:
        yield _MockPartition(self._name)

    @property
    def name(self) -> str:
        return self._name

    @property
    def cursor_field(self) -> Optional[str]:
        raise NotImplementedError

    def check_availability(self) -> StreamAvailability:
        if self._available:
            return StreamAvailable()
        else:
            return StreamUnavailable("stream is unavailable")

    def get_json_schema(self) -> Mapping[str, Any]:
        return self._json_schema

    def as_airbyte_stream(self) -> AirbyteStream:
        return AirbyteStream(name=self.name, json_schema=self.get_json_schema(), supported_sync_modes=[SyncMode.full_refresh])

    def log_stream_sync_configuration(self) -> None:
        raise NotImplementedError

    @property
    def cursor(self) -> Cursor:
        return FinalStateCursor(stream_name=self._name, stream_namespace=None, message_repository=self._message_repository)


class _MockPartition(Partition):
    def __init__(self, name: str):
        self._name = name
        self._closed = False

    def read(self) -> Iterable[Record]:
        yield from [Record({"key": "value"}, self._name)]

    def to_slice(self) -> Optional[Mapping[str, Any]]:
        return {}

    def stream_name(self) -> str:
        return self._name

    def close(self) -> None:
        self._closed = True

    def is_closed(self) -> bool:
        return self._closed

    def __hash__(self) -> int:
        return hash(self._name)


def test_concurrent_source_reading_from_no_streams():
    message_repository = InMemoryMessageRepository()
    stream = _MockStream("my_stream", message_repository,False, {})
    source = _MockSource(message_repository=message_repository)
    messages = []
    for m in source.read([stream]):
        messages.append(m)
